// ############################################################################################
//  DefiniteAreaFractals.cc
//  Implementation of CDefiniteAreaFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Pemrograman Fraktal Resolusi Tinggi untuk Kartu Trident SVGA
//    Ir. P. Insap Santosa, M.Sc. & Ir. R. Tory K.
//    Penerbit ANDI Yogyakarta
//    Indonesia, 1997
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "DefiniteAreaFractals.h"

// C ++ headers
#include<complex>


namespace vge {

//
// The base of all complex-number-based fractals
//
CCNBFractal::CCNBFractal()
: _bCWInitialized(false), _bCSInitialized(false)
{
  memset(&_tW, 0, sizeof(_tW));
}

CCNBFractal::~CCNBFractal() throw()
{
}

void CCNBFractal::SelectWindow(float fX1, float fY1, float fX2, float fY2)
{
  if(fX1 < fX2) {
    _tW.fX1 = fX1;
    _tW.fX2 = fX2;
  }
  else {
    _tW.fX1 = fX2;
    _tW.fX2 = fX1;
  }
  if(fY1 < fY2) {
    _tW.fY1 = fY1;
    _tW.fY2 = fY2;
  }
  else {
    _tW.fY1 = fY2;
    _tW.fY2 = fY1;
  }
  _bCWInitialized = true;
}

void CCNBFractal::SelectSeed(float fPv, float fQv)
{
  _tW.fPv = fPv;
  _tW.fQv = fQv;
  _bCSInitialized = true;
}

uint32_t CCNBFractal::EquMaxIteration() { return(INDEFINITE_AREA_FRACTAL_DEFAULT_ITERATION); }

void CCNBFractal::UpdateColor(uint32_t un32MaxIter, uint32_t un32Iter, float fColorMulFactor)
{
  if((un32Iter == 0 ) || (un32Iter >= un32MaxIter))
    SetColor(0, 0, 0);
  else {
    int32_t n32R = int32_t((un32MaxIter - un32Iter) * fColorMulFactor * 0.5f);
    int32_t n32G = int32_t(               un32Iter  * fColorMulFactor);
    int32_t n32B = int32_t((un32MaxIter - un32Iter) * fColorMulFactor);
    SetColor((n32R < 0) ? 0 : ((n32R > int32_t(MAX_COLOR_VALUE)) ? MAX_COLOR_VALUE : n32R),
             (n32G < 0) ? 0 : ((n32G > int32_t(MAX_COLOR_VALUE)) ? MAX_COLOR_VALUE : n32G),
             (n32B < 0) ? 0 : ((n32B > int32_t(MAX_COLOR_VALUE)) ? MAX_COLOR_VALUE : n32B));
  }
}

void CCNBFractal::DrawImpl(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2, uint32_t un32MaxIter)
{
  // Initialize some parameters (if haven't)
  if(!_bCWInitialized) SelectDefaultWindow();
  if(!_bCSInitialized) SelectDefaultSeed();

  // Perparation
  int32_t  n32DeltaX = n32X2 - n32X1;
  int32_t  n32DeltaY = n32Y2 - n32Y1;
  int32_t  n32H      = _pcsIBuild->GetHeight() - n32Y1 - n32DeltaY;
  float    fDeltaP   = float(_tW.fX2 - _tW.fX1) / float(n32DeltaX + 1);
  float    fDeltaQ   = float(_tW.fY2 - _tW.fY1) / float(n32DeltaY + 1);
  float    fCMFactor = float(MAX_COLOR_VALUE) / float(un32MaxIter - 1);

  // Draw the fractal
  for(int32_t n32CntY = 0; n32CntY <= n32DeltaY; n32CntY++) {
    // Call the callback first
    DrawingProgressCallback(n32DeltaX, n32DeltaX, n32DeltaY, n32CntY);
    // Calculate the Y coordinate
    float fY = _tW.fY1 + fDeltaQ * n32CntY;
    // Draw the row
    for(int32_t n32CntX = 0; n32CntX <= n32DeltaX; n32CntX++) {
      // Calculate the X coordinate
      float fX = _tW.fX1 + fDeltaP * n32CntX;
      // Iterate the equation
      uint32_t un32Count = IterateEqu(fX, fY, un32MaxIter);
      // Calculate the color
      UpdateColor(un32MaxIter - 1, un32Count, fCMFactor);
      // Draw the pixel
      PutPixel(n32X1 + n32CntX, n32H + n32CntY);
    }

  }
}

void CCNBFractal::Draw(CImageBuilder *pcsDstIBuild, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Copy the target image builder
  _pcsIBuild = pcsDstIBuild;

  // Draw the fractal
  DrawImpl(n32X1, n32Y1, n32X2, n32Y2, EquMaxIteration());
}

// Drawing callback functions
// NOTE: * Override this function to display/control the drawing progress.
//       * Returning 'false' will stop the drawing.
bool CCNBFractal::DrawingProgressCallback(int32_t n32MaxX, int32_t n32CurX, int32_t n32MaxY, int32_t n32CurY)
{
  UNUSED_PARAMS_4(n32MaxX, n32CurX, n32MaxY, n32CurY);
  return(true);
}


//
// The base of all transcendental-function-based fractals
//
uint32_t CTFBFractal::IterateEqu(float fX, float fY, uint32_t un32MaxIter)
{
  // Iterate the equation : z(n) = f[z(n-1), c]
  //                       with : z(0) = 0
  //                              c    = x + iy       (from fX and fY parameters)
  //
  //                 Exit point : when |z(n)| >= 1000.0
  complex<float> tZ(0, 0);
  complex<float> tC(fX, fY);
  uint32_t un32Cnt = 0;
  while((un32Cnt < un32MaxIter) && (norm(tZ) < 1000.0f)) {
    CalculateEqu(tZ, tC);
    un32Cnt++;
  }
  return(un32Cnt);
}


//
// The base of all pixel-color-interpolator fractals
//
uint8_t CPCIFractal::RandomFactor(int32_t n32Distance)
{
  UNUSED_PARAM(n32Distance);
  return(0);
}

void CPCIFractal::CalculateColor(uint8_t &un8Rx, uint8_t &un8Gx, uint8_t &un8Bx, uint8_t un8R1, uint8_t un8G1, uint8_t un8B1, uint8_t un8R2, uint8_t un8G2, uint8_t un8B2)
{
  un8Rx = (un8R1 + un8R2) / 2;
  un8Gx = (un8G1 + un8G2) / 2;
  un8Bx = (un8B1 + un8B2) / 2;
}

void CPCIFractal::RandomizeColor(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B, uint8_t un8Factor)
{
  if(un8Factor) {
    int8_t nF1 = int8_t((random() / float(RAND_MAX) * 2.0f * un8Factor) - un8Factor);
    int8_t nF2 = int8_t((random() / float(RAND_MAX) * 2.0f * un8Factor) - un8Factor);
    int8_t nF3 = int8_t((random() / float(RAND_MAX) * 2.0f * un8Factor) - un8Factor);
    int8_t nRf = int8_t(un8R) + nF1;
    int8_t nGf = int8_t(un8G) + nF2;
    int8_t nBf = int8_t(un8B) + nF3;
    un8R = (nRf < 0) ? 0 : ((nRf > MAX_COLOR_VALUE) ? MAX_COLOR_VALUE : nRf);
    un8G = (nGf < 0) ? 0 : ((nGf > MAX_COLOR_VALUE) ? MAX_COLOR_VALUE : nGf);
    un8B = (nBf < 0) ? 0 : ((nBf > MAX_COLOR_VALUE) ? MAX_COLOR_VALUE : nBf);
  }
}

bool CPCIFractal::DrawImpl(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2, bool bDrawCenterPixel, bool bFirstCall)
{
  // Convert the n32Y1 and n32Y2 from standard screen coordinate to cartesian space
  if(bFirstCall) {
    n32Y1 = _pcsIBuild->GetHeight() - n32Y1;
    n32Y2 = _pcsIBuild->GetHeight() - n32Y2;
  }

  // Check
  int32_t n32Xd = n32X2 - n32X1;
  int32_t n32Yd = n32Y2 - n32Y1;
  n32Xd = (n32Xd < 0) ? (-n32Xd) : n32Xd;
  n32Yd = (n32Yd < 0) ? (-n32Yd) : n32Yd;
  if((n32Xd < 2) && (n32Yd < 2)) return(true);

  // Call the callback first
  if(bFirstCall) {
    if(!DrawingProgressCallback(n32X1, n32Y1, n32X2, n32Y2)) return(true);
  }

  // Get the pixels
  uint8_t un8R1, un8G1, un8B1;
  uint8_t un8R2, un8G2, un8B2;
  uint8_t un8R3, un8G3, un8B3;
  uint8_t un8R4, un8G4, un8B4;
  GetPixel(n32X1, n32Y1, un8R1, un8G1, un8B1); // top-left
  GetPixel(n32X2, n32Y1, un8R2, un8G2, un8B2); // top-right
  GetPixel(n32X1, n32Y2, un8R3, un8G3, un8B3); // bottom-left
  GetPixel(n32X2, n32Y2, un8R4, un8G4, un8B4); // bottom-right

  // Calculate new colors
  uint8_t un8R1x, un8G1x, un8B1x;
  uint8_t un8R2x, un8G2x, un8B2x;
  uint8_t un8R3x, un8G3x, un8B3x;
  uint8_t un8R4x, un8G4x, un8B4x;
  uint8_t un8R5x, un8G5x, un8B5x;
  CalculateColor(un8R1x, un8G1x, un8B1x, un8R1, un8G1, un8B1, un8R2, un8G2, un8B2); // top-center
  CalculateColor(un8R2x, un8G2x, un8B2x, un8R2, un8G2, un8B2, un8R4, un8G4, un8B4); // center-right
  CalculateColor(un8R3x, un8G3x, un8B3x, un8R3, un8G3, un8B3, un8R4, un8G4, un8B4); // bottom-center
  CalculateColor(un8R4x, un8G4x, un8B4x, un8R1, un8G1, un8B1, un8R3, un8G3, un8B3); // center-left
  RandomizeColor(un8R1x, un8G1x, un8B1x, RandomFactor(n32X2 - n32X1));
  RandomizeColor(un8R2x, un8G2x, un8B2x, RandomFactor(n32Y2 - n32Y1));
  RandomizeColor(un8R3x, un8G3x, un8B3x, RandomFactor(n32X2 - n32X1));
  RandomizeColor(un8R4x, un8G4x, un8B4x, RandomFactor(n32Y2 - n32Y1));
  if(bDrawCenterPixel) {
    un8R5x = (un8R1x + un8R2x + un8R3x + un8R4x) / 4;
    un8G5x = (un8G1x + un8G2x + un8G3x + un8G4x) / 4;
    un8B5x = (un8B1x + un8B2x + un8B3x + un8B4x) / 4;
  }
  else {
      un8R5x = un8G5x = un8B5x = 0;
  }

  // Calculate centers
  int32_t n32Xc = (n32X1 + n32X2) / 2;
  int32_t n32Yc = (n32Y1 + n32Y2) / 2;

  // Put the pixels
  SetPixel(n32Xc, n32Y1, un8R1x, un8G1x, un8B1x); // top-center
  SetPixel(n32X2, n32Yc, un8R2x, un8G2x, un8B2x); // center-right
  SetPixel(n32Xc, n32Y2, un8R3x, un8G3x, un8B3x); // bottom-center
  SetPixel(n32X1, n32Yc, un8R4x, un8G4x, un8B4x); // center-left
  if(bDrawCenterPixel)
    SetPixel(n32Xc, n32Yc, un8R5x, un8G5x, un8B5x); // center-center

  // Draw the next blocks recursively
  if(bFirstCall) {
    if(!DrawingProgressCallback(n32X1, n32Y1, n32Xc, n32Yc)) return(true);
    DrawImpl(n32X1, n32Y1, n32Xc, n32Yc, bDrawCenterPixel); // top-left
    if(!DrawingProgressCallback(n32Xc, n32Y1, n32X2, n32Yc)) return(true);
    DrawImpl(n32Xc, n32Y1, n32X2, n32Yc, bDrawCenterPixel); // top-right
    if(!DrawingProgressCallback(n32X1, n32Yc, n32Xc, n32Y2)) return(true);
    DrawImpl(n32X1, n32Yc, n32Xc, n32Y2, bDrawCenterPixel); // bottom-left
    if(!DrawingProgressCallback(n32Xc, n32Yc, n32X2, n32Y2)) return(true);
    DrawImpl(n32Xc, n32Yc, n32X2, n32Y2, bDrawCenterPixel); // bottom-right
  }
  else {
    DrawImpl(n32X1, n32Y1, n32Xc, n32Yc, bDrawCenterPixel); // top-left
    DrawImpl(n32Xc, n32Y1, n32X2, n32Yc, bDrawCenterPixel); // top-right
    DrawImpl(n32X1, n32Yc, n32Xc, n32Y2, bDrawCenterPixel); // bottom-left
    DrawImpl(n32Xc, n32Yc, n32X2, n32Y2, bDrawCenterPixel); // bottom-right
  }
  return(true);
}

void CPCIFractal::Draw(CImageBuilder *pcsDstIBuild, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Copy the target image builder
  _pcsIBuild = pcsDstIBuild;

  // Draw the fractal
  DrawSeed(n32X1, n32Y1, n32X2, n32Y2);
  DrawImpl(n32X1, n32Y1, n32X2, n32Y2, DrawCenterPixel(), true);
}

bool CPCIFractal::DrawingProgressCallback(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  UNUSED_PARAMS_4(n32X1, n32Y1, n32X2, n32Y2);
  return(true);
}

} // namespace vge
