// ############################################################################################
//  3DEnvironment.cc
//  Implementation of C3DEnv class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../VideoEnvironment/VideoEnvironment.h"
#include "3DEnvironment.h"
#include "FixedPointConsts.h"


namespace vge {

// The default render size
const float _XRENDER_  = 320.0f;
const float _YRENDER_  = 240.0f;

// The default scalling factor
const float _XSCALE_  =  300.0f;
const float _YSCALE_  = -300.0f;


//
// The 3D environment class
//
C3DEnv *C3DEnv::_active3DEnv = NULL;

C3DEnv::C3DEnv()
: _tRB      (NULL), _n32FCount(0),
  _pun32RFLT(NULL), _pun32GFLT(NULL), _pun32BFLT(NULL),
  _pun32RFCT(NULL), _pun32GFCT(NULL), _pun32BFCT(NULL),
  _pun8TRFCT(NULL), _pun8TGFCT(NULL), _pun8TBFCT(NULL),
  _pun8TRFLT(NULL), _pun8TGFLT(NULL), _pun8TBFLT(NULL)
{
  // Allocate memory for fog tables
  _pun32RFLT = _nothrow_new uint32_t[32*32*32];
  _pun32GFLT = _nothrow_new uint32_t[32*32*32];
  _pun32BFLT = _nothrow_new uint32_t[32*32*32];
  _pun32RFCT = _nothrow_new uint32_t[32*32];
  _pun32GFCT = _nothrow_new uint32_t[32*32];
  _pun32BFCT = _nothrow_new uint32_t[32*32];
  _pun8TRFCT = _nothrow_new  uint8_t[32*32*32];
  _pun8TGFCT = _nothrow_new  uint8_t[32*32*32];
  _pun8TBFCT = _nothrow_new  uint8_t[32*32*32];
  _pun8TRFLT = _nothrow_new  uint8_t[32*32*32*32];
  _pun8TGFLT = _nothrow_new  uint8_t[32*32*32*32];
  _pun8TBFLT = _nothrow_new  uint8_t[32*32*32*32];
  if(!_pun32RFLT || !_pun32GFLT || !_pun32BFLT ||
     !_pun32RFCT || !_pun32GFCT || !_pun32BFCT ||
     !_pun8TRFCT || !_pun8TGFCT || !_pun8TBFCT ||
     !_pun8TRFLT || !_pun8TGFLT || !_pun8TBFLT) {
    delete [] _pun32RFLT; delete [] _pun32GFLT; delete [] _pun32BFLT;
    delete [] _pun32RFCT; delete [] _pun32GFCT; delete [] _pun32BFCT;
    delete [] _pun8TRFCT; delete [] _pun8TGFCT; delete [] _pun8TBFCT;
    delete [] _pun8TRFLT; delete [] _pun8TGFLT; delete [] _pun8TBFLT;
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for fog tables");
  }
}

C3DEnv::~C3DEnv() throw()
{
  delete [] _pun32RFLT; delete [] _pun32GFLT; delete [] _pun32BFLT;
  delete [] _pun32RFCT; delete [] _pun32GFCT; delete [] _pun32BFCT;
  delete [] _pun8TRFCT; delete [] _pun8TGFCT; delete [] _pun8TBFCT;
  delete [] _pun8TRFLT; delete [] _pun8TGFLT; delete [] _pun8TBFLT;
}

void C3DEnv::Initialize(CVideoEnv *pcsVVideoEnv)
{
  // Check
  vassert(pcsVVideoEnv->_bAllInitialized,
          "The given CVideoEnv must be completely initialized before initializing C3DEnv");

  // Copy variables
  _tRB = &(pcsVVideoEnv->_tRB);

  // Initialize the default ambient light
  _tDefALight.un8R = _tDefALight.un8G = _tDefALight.un8B = MAX_COLOR_VALUE / 2;

  // Initialize the default omni light
  _tDefOLight.un8R = _tDefOLight.un8G = _tDefOLight.un8B = MAX_COLOR_VALUE;
  _tDefOLight.fX   = _tDefOLight.fY   = 1000;
  _tDefOLight.fZ   = -1000;

  // Initialize the viewport
  SetZClipping();
  SetViewport();

  // Initialize some Z-buffer constants
  _tRB->n32ZTransInc = 1 << _ZSTEP_PREC_;
  _tRB->n32ZWait    = int32_t(pow(2.0f, int32_t(31 - _ZSTEP_PREC_))) - 1;

  // Initialize LUTs
  xprv::_VInitSinCosTable_();
  xprv::_VInitLightingTable_();
  SetFogState(16, 16, 16, 8);

  // If this is the first 3D environment then set it as the active one
  if(C3DEnv::Active3DEnv() == NULL) SetAsTheActiveEnv();
}

void C3DEnv::SetAsTheActiveEnv()
{
  C3DEnv::_active3DEnv        = this;
  xprv::_pun32RFogLightTable_ = _pun32RFLT;
  xprv::_pun32GFogLightTable_ = _pun32GFLT;
  xprv::_pun32BFogLightTable_ = _pun32BFLT;
  xprv::_pun32RFogColorTable_ = _pun32RFCT;
  xprv::_pun32GFogColorTable_ = _pun32GFCT;
  xprv::_pun32BFogColorTable_ = _pun32BFCT;
  xprv::_pun8TRFogColorTable_ = _pun8TRFCT;
  xprv::_pun8TGFogColorTable_ = _pun8TGFCT;
  xprv::_pun8TBFogColorTable_ = _pun8TBFCT;
  xprv::_pun8TRFogLightTable_ = _pun8TRFLT;
  xprv::_pun8TGFogLightTable_ = _pun8TGFLT;
  xprv::_pun8TBFogLightTable_ = _pun8TBFLT;
}

void C3DEnv::SetFogState(uint8_t un8FogColorR, uint8_t un8FogColorG, uint8_t un8FogColorB, uint8_t un8FogDepth)
{
  // Copy variables
  _un8FogR = un8FogColorR;
  _un8FogG = un8FogColorG;
  _un8FogB = un8FogColorB;

  // Set depth
  if(un8FogDepth <= 1)
    _un32FogShift = 15;
  else if(un8FogDepth >= 15)
    _un32FogShift = 1;
  else
    _un32FogShift = 16 - un8FogDepth;

  // Generate the LUTS
  for(uint32_t un32CF = 0; un32CF < 32; un32CF++) {
    for(uint32_t un32C1 = 0; un32C1 < 32; un32C1++) {
      for(uint32_t un32C2 = 0; un32C2 < 32; un32C2++) {
        // Generate the fogged-lighting LUTs
        uint32_t un32T = ((un32C1 * un32C2) / 31) & 0x1F;
        uint32_t un32R = ((un32CF * uint32_t(un8FogColorR)) + ((31 - un32CF) * un32T)) / 31;
        uint32_t un32G = ((un32CF * uint32_t(un8FogColorG)) + ((31 - un32CF) * un32T)) / 31;
        uint32_t un32B = ((un32CF * uint32_t(un8FogColorB)) + ((31 - un32CF) * un32T)) / 31;
        _pun32RFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32R << 19;
        _pun32GFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32G << 11;
        _pun32BFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32B << 3;
        // Generate the fogged-lighting LUTs with transparency
        for(uint32_t un32OC = 0; un32OC < 32; un32OC++) {
          _pun8TRFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = uint8_t((un32OC + un32R) / 2) << 3;
          _pun8TGFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = uint8_t((un32OC + un32G) / 2) << 3;
          _pun8TBFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = uint8_t((un32OC + un32B) / 2) << 3;
        }
      }
      // Generate the fogged-color LUTs
      uint32_t un32R = ((un32CF * uint32_t(un8FogColorR)) + ((31 - un32CF) * un32C1)) / 31;
      uint32_t un32G = ((un32CF * uint32_t(un8FogColorG)) + ((31 - un32CF) * un32C1)) / 31;
      uint32_t un32B = ((un32CF * uint32_t(un8FogColorB)) + ((31 - un32CF) * un32C1)) / 31;
      _pun32RFCT[(un32CF << 5) | un32C1] = un32R << 19;
      _pun32GFCT[(un32CF << 5) | un32C1] = un32G << 11;
      _pun32BFCT[(un32CF << 5) | un32C1] = un32B << 3;
      // Generate the fogged-color LUTs with transparency
      for(uint32_t un32OC = 0; un32OC < 32; un32OC++) {
        _pun8TRFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = uint8_t((un32R + un32OC) / 2) << 3;
        _pun8TGFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = uint8_t((un32G + un32OC) / 2) << 3;
        _pun8TBFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = uint8_t((un32B + un32OC) / 2) << 3;
      }
    }
  }
}

void C3DEnv::SetZClipping(uint32_t un32MinZ, uint32_t un32MaxZ)
{
  if(un32MinZ ==  0) VError(VERR_InvalidParameter, _NFL_, "The minimum Z clipping value must be > 0");
  if(un32MinZ  < 50) VWarning(_NFL_, "Setting the Z clipping value with any value < 50 may cause strange rendering artifact");
  
  if(un32MaxZ <= un32MinZ) VError(VERR_InvalidParameter, _NFL_, "The maximum Z clipping value must be > the minimum one");

  _un32MinZ = un32MinZ;
  _un32MaxZ = un32MaxZ;
}

void C3DEnv::SetViewport(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Calculate the clipping region
  if(n32X1 >= 0)
    _n32MinX = uint32_t(n32X1);
  else
    _n32MinX = 2;
  if(n32Y1 >= 0)
    _n32MinY = uint32_t(n32Y1);
  else
    _n32MinY = 2;
  if(n32X2 >= 0)
    _n32MaxX = uint32_t(n32X2);
  else
    _n32MaxX = int32_t(_tRB->un32Width - 2);
  if(n32Y2 >= 0)
    _n32MaxY = uint32_t(n32Y2);
  else
    _n32MaxY = int32_t(_tRB->un32Height - 2);

  // Validate the clipping region
  if(_n32MinX < 0) _n32MinX = 0;
  if(_n32MinY < 0) _n32MinY = 0;
  if(uint32_t(_n32MaxX) > _tRB->un32Width)  _n32MaxX = int32_t(_tRB->un32Width);
  if(uint32_t(_n32MaxY) > _tRB->un32Height) _n32MaxY = int32_t(_tRB->un32Height);

  // Calculate projection shift and scale
  _fXShift  = float(_n32MinX) + float(_n32MaxX - _n32MinX) / 2.0f;
  _fYShift  = float(_n32MinY) + float(_n32MaxY - _n32MinY) / 2.0f;
  _fXScale  = _XSCALE_ * float(_n32MaxX - _n32MinX) / _XRENDER_;
  _fYScale  = _YSCALE_ * float(_n32MaxY - _n32MinY) / _YRENDER_;
  _fXScale4 = _fXScale / 4.0f;
  _fYScale4 = _fYScale / 4.0f;
}

void C3DEnv::PutImage(const CImage &csImagePut, uint32_t un32X, int32_t un32Y)
{
  // Preparation
  const uint32_t *pun32STmp = reinterpret_cast<const uint32_t*>(csImagePut._pun8Buffer);
        uint32_t *pun32DTmp = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[(un32Y * _tRB->un32Width + un32X) << 2]);
        uint32_t  un32Add   = _tRB->un32Width - csImagePut._un32Width;

  // Blit
  for(uint32_t un32Yc = 0; un32Yc < csImagePut._un32Height; un32Yc++) {
    for(uint32_t un32Xc = 0; un32Xc < csImagePut._un32Width; un32Xc++)
      *pun32DTmp++ = *pun32STmp++;
    pun32DTmp += un32Add;
  }
}

void C3DEnv::PutImage(const CImage &csImagePut, uint32_t un32X, int32_t un32Y, XColor15 xExcludedColorVRGB)
{
  // Preparation
  const uint32_t *pun32STmp = reinterpret_cast<const uint32_t*>(csImagePut._pun8Buffer);
        uint32_t *pun32DTmp = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[(un32Y * _tRB->un32Width + un32X) << 2]);
        uint32_t  un32Add   = _tRB->un32Width - csImagePut._un32Width;
  if(xExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL) xExcludedColorVRGB = *pun32STmp;

  // Blit
  for(uint32_t un32Yc = 0; un32Yc < csImagePut._un32Height; un32Yc++) {
    for(uint32_t un32Xc = 0; un32Xc < csImagePut._un32Width; un32Xc++) {
      if(*pun32STmp != xExcludedColorVRGB)
        *pun32DTmp++ = *pun32STmp++;
      else {
        pun32DTmp++;
        pun32STmp++;
      }
    }
    pun32DTmp += un32Add;
  }
}

void C3DEnv::ClipPutImage(const CImage &csImagePut, int32_t n32X, int32_t n32Y)
{
  // Preparation
  const uint32_t *pun32STmp1, *pun32STmp2;
        uint32_t *pun32DTmp1, *pun32DTmp2;
        uint32_t  un32Ws, un32Hs;
  if(n32Y < 0) {
    if(n32X < 0) {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[(uint32_t(-n32Y) * csImagePut._un32Width + uint32_t(-n32X)) << 2]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(_tRB->pun8Buffer);
      un32Ws     = uint32_t(int32_t(csImagePut._un32Width) + n32X);
      n32X       = 0;
    }
    else {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[uint32_t(-n32Y) * csImagePut._un32SizeSL]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[n32X << 2]);
      un32Ws     = csImagePut._un32Width;
    }
    un32Hs = uint32_t(int32_t(csImagePut._un32Height) + n32Y);
    n32Y   = 0;
  }
  else {
    if(n32X < 0) {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[uint32_t(-n32X) << 2]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[n32Y * _tRB->un32SizeSL]);
      un32Ws     = uint32_t(int32_t(csImagePut._un32Width) + n32X);
      n32X       = 0;
    }
    else {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(csImagePut._pun8Buffer);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[(n32Y * _tRB->un32Width + n32X) << 2]);
      un32Ws     = csImagePut._un32Width;
    }
    un32Hs = csImagePut._un32Height;
  }

  // Blit
  for(uint32_t un32Yc = 0; un32Yc < un32Hs; un32Yc++) {
    if((un32Yc + n32Y) >= _tRB->un32Height) break;
    pun32STmp2 = pun32STmp1;
    pun32DTmp2 = pun32DTmp1;
    for(uint32_t un32Xc = 0; un32Xc < un32Ws; un32Xc++) {
      if((un32Xc + n32X) >= _tRB->un32Width) break;
      *pun32DTmp2++ = *pun32STmp2++;
    }
    pun32STmp1 += csImagePut._un32Width;
    pun32DTmp1 += _tRB->un32Width;
  }
}

void C3DEnv::ClipPutImage(const CImage &csImagePut, int32_t n32X, int32_t n32Y, XColor15 xExcludedColorVRGB)
{
  // Preparation
  const uint32_t *pun32STmp1, *pun32STmp2;
        uint32_t *pun32DTmp1, *pun32DTmp2;
        uint32_t  un32Ws, un32Hs;
  if(n32Y < 0) {
    if(n32X < 0) {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[(uint32_t(-n32Y) * csImagePut._un32Width + uint32_t(-n32X)) << 2]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(_tRB->pun8Buffer);
      un32Ws     = uint32_t(int32_t(csImagePut._un32Width) + n32X);
      n32X       = 0;
    }
    else {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[uint32_t(-n32Y) * csImagePut._un32SizeSL]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[n32X << 2]);
      un32Ws     = csImagePut._un32Width;
    }
    un32Hs = uint32_t(int32_t(csImagePut._un32Height) + n32Y);
    n32Y   = 0;
  }
  else {
    if(n32X < 0) {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(&csImagePut._pun8Buffer[uint32_t(-n32X) << 2]);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[n32Y * _tRB->un32SizeSL]);
      un32Ws     = uint32_t(int32_t(csImagePut._un32Width) + n32X);
      n32X       = 0;
    }
    else {
      pun32STmp1 = reinterpret_cast<const uint32_t*>(csImagePut._pun8Buffer);
      pun32DTmp1 = reinterpret_cast<      uint32_t*>(&_tRB->pun8Buffer[(n32Y * _tRB->un32Width + n32X) << 2]);
      un32Ws     = csImagePut._un32Width;
    }
    un32Hs = csImagePut._un32Height;
  }
  if(xExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL)
    xExcludedColorVRGB = *(reinterpret_cast<const uint32_t*>(csImagePut._pun8Buffer));

  // Blit
  for(uint32_t un32Yc = 0; un32Yc < un32Hs; un32Yc++) {
    if((un32Yc + n32Y) >= _tRB->un32Height) break;
    pun32STmp2 = pun32STmp1;
    pun32DTmp2 = pun32DTmp1;
    for(uint32_t un32Xc = 0; un32Xc < un32Ws; un32Xc++) {
      if((un32Xc + n32X) >= _tRB->un32Width) break;
      if(*pun32STmp2 != xExcludedColorVRGB)
        *pun32DTmp2++ = *pun32STmp2++;
      else {
        pun32DTmp2++;
        pun32STmp2++;
      }
    }
    pun32STmp1 += csImagePut._un32Width;
    pun32DTmp1 += _tRB->un32Width;
  }
}


//
// Calculate the angle (cosine value) between the panel's normal and the light's vector
//
float VCalcLightCosValue(const float *pfPanelNormalXYZ4, const float *pfPanelPointXYZ4, const float *pfLightPosXYZ4)
{
  float afT[4], fMag, fCos;

  // Create the normalized light vector from the panel's point to the omni light position
  afT[0] = pfLightPosXYZ4[0] - pfPanelPointXYZ4[0];
  afT[1] = pfLightPosXYZ4[1] - pfPanelPointXYZ4[1];
  afT[2] = pfLightPosXYZ4[2] - pfPanelPointXYZ4[2];
  fMag = sqrt(afT[0] * afT[0] + afT[1] * afT[1] + afT[2] * afT[2]);
  afT[0] /= fMag; afT[1] /= fMag; afT[2] /= fMag;

  // Return the angle (cosine value) between panel's normal and the light vector
  fCos= pfPanelNormalXYZ4[0] * afT[0] + pfPanelNormalXYZ4[1] * afT[1] + pfPanelNormalXYZ4[2] * afT[2];
  fCos = (fCos < 0) ? 0 : fCos;
  return(fCos);
}

} // namespace vge
