/**
 * \file HierarchicalNode.h
 * <hr>
 * <div style="text-align:center;"><b>!!! WARNING !!!</b></div>
 * \n
 * NEVER delete child nodes manually because all children will be deleted automatically
 * when the parent is deleted.
 * \n\n
 * If you need to preserve a child node then change its parent to NULL before deleting
 * the parent.
 * <hr>
 */
// ############################################################################################
//  HierarchicalNode.h
//  Definition & implementation of CHNode class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___HIERARCHICAL_NODE_H_
#define _3D_ENVIRONMENT___HIERARCHICAL_NODE_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"
#include "../System/DataSinkSource.h"
#include "Matrix3D.h"

namespace vge {

  // ##########################################################################################
  // #################################  W A R N I N G  ! ! !  #################################
  // ##########################################################################################
  // ###### - NEVER delete child nodes manually because all children will be deleted     ######
  // ######   automatically when the parent is deleted.                                  ######
  // ###### - If you need to preserve a child node then change its parent to NULL before ######
  // ######   deleting the parent.                                                       ######
  // ##########################################################################################
  // #################################  W A R N I N G  ! ! !  #################################
  // ##########################################################################################


  // Maximum children allowed
  const uint32_t _MAX_CHILD_ = uint32_t(64);


  //!
  //! \brief The hierarchical node class
  //!
  //! \api pretty stable
  class VGE_EXPORT CHNode : VCLASS_UNCOPYABLE {
    public:
      CHNode();
      ~CHNode() throw();

      // Hierarchy constructions
      //!
      //! Attach this node to the given parent (will detach from any old parent).
      void SetParent(CHNode *pcsNewParent);
      //!
      //! Attach this node to the given parent (will detach from any old parent).
      inline void SetParent(CHNode &csNewParent)
      { SetParent(&csNewParent); }
      //!
      //! Create a new child node with the given user ID (optional).
      CHNode *AddChild(uint32_t un32UserID = 0);
      //!
      //! Delete the given child node instance.
      void DelChild(CHNode *pcsChild);
      //!
      //! Delete all the child node instances of this node.
      void DelAllChild();

      // Information functions
      //!
      //! Get the parent node of this node.
      inline CHNode *GetParent()
      { return(_pcsParent); }
      //!
      //! Get the parent node of this node.
      inline const CHNode *GetParent() const
      { return(_pcsParent); }
      //!
      //! Set the user ID for this node.
      inline void SetID(uint32_t un32NewUserID = 0)
      { _un32UserID = un32NewUserID; }
      //!
      //! Get the user ID for this node.
      inline uint32_t GetID() const
      { return(_un32UserID); }
      //!
      //! Find any child node which has the given user ID.
      //! \param[in] un32UserID         The user ID to be searched, if multiple
      //!                               children having the same ID exist, the
      //!                               first found one will be returned.
      //! \param[in] bRecursiveSubChild Shall the search be done in this node only
      //!                               or in the whole tree.
      CHNode *FindChild(uint32_t un32UserID, bool bRecursiveSubChild = true) const;

      // Dynamic matrix access
      //!
      //! Returns the merged transformation matrix for this node.
      inline CMatrix3D &Matrix() { return(_tM.csTot); }
      //!
      //! Returns the merged transformation matrix for this node.
      inline const CMatrix3D &Matrix() const { return(_tM.csTot); }
      //!
      //! Update the merged transformation matrix of this node and
      //! all of its child nodes if the "bUpdateChildren" parameter
      //! is set to true.
      //! \n\n
      //! The operations are as follow:\n<PRE>
      //!     |N_Mrg| = |N_Stc|
      //!     |N_Mrg| = |N_Mrg| x |N_Scl|
      //!     |N_Mrg| = |N_Mrg| x |N_Rot|
      //!     |N_Mrg| = |N_Mrg| x |N_Trn|
      //!     |N_Mrg| = |N_Mrg| x |P_Mrg|
      //! </PRE>
      //! where:\n<PRE>
      //!     |N_Mrg| : This node's merged matrix
      //!     |N_Stc| : This node's static matrix
      //!     |N_Scl| : This node's dynamic scalling matrix
      //!     |N_Rot| : This node's dynamic rotation matrix
      //!     |N_Trn| : This node's dynamic translation matrix
      //!     |P_Mrg| : Parent node's merged matrix (if this node has parent)
      //! </PRE>
      void UpdateMatrix(bool bUpdateChildren = true);

      // Static matrix operations
      //!
      //! Make the node's static matrix become an identity matrix.
      inline void StcIdentity();
      //!
      //! Shear the node's static matrix by the given values.
      inline void StcShear(float fX, float fY);
      //!
      //! Scale the node's static matrix by the given values.
      inline void StcScale(float fX, float fY, float fZ);
      //!
      //! Roatate the node's static matrix by the given values.
      inline void StcRotate(int32_t n32X, int32_t n32Y, int32_t n32Z);
      //!
      //! Translate the node's static matrix by the given values.
      inline void StcTranslate(float fX, float fY, float fZ);
      //!
      //! Copy the given array to the node's static matrix.
      inline void StcSet(float const aafNewMatrix[4][4]);
      //!
      //! Copy the given matrix to the node's static matrix.
      inline void StcSet(CMatrix3D &csVNewMatrix);

      // Dynamic matrix operations
      //!
      //! Make the node's dynamic matrix become an identity matrix.
      inline void DynSetIdentity();
      //!
      //! Set the node's dynamic scalling matrix with the given values.
      inline void DynSetScalling(float fX, float fY, float fZ);
      //!
      //! Set the node's dynamic rotation matrix with the given values.
      inline void DynSetRotation(int32_t n32X, int32_t n32Y, int32_t n32Z);
      //!
      //! Set the node's dynamic position matrix with the given values.
      inline void DynSetPosition(float fX, float fY, float fZ);
      //!
      //! Scale the node's dynamic scalling matrix by the given values.
      inline void DynScale(float fX, float fY, float fZ);
      //!
      //! Rotate the node's dynamic rotation matrix by the given values.
      inline void DynRotate(int32_t n32X, int32_t n32Y, int32_t n32Z);
      //!
      //! Rotate the node's dynamic rotation matrix by the given value
      //! (X axis only).
      inline void DynRotateX(int32_t n32X);
      //!
      //! Rotate the node's dynamic rotation matrix by the given value
      //! (Y axis only).
      inline void DynRotateY(int32_t n32Y);
      //!
      //! Rotate the node's dynamic rotation matrix by the given value
      //! (Z axis only).
      inline void DynRotateZ(int32_t n32Z);
      //!
      //! Translate the node's dynamic position matrix by the given values.
      inline void DynTranslate(float fX, float fY, float fZ);

      // I/O functions
      //!
      //! Save to VortexGE's native VHN format.
      //! \param[in] sink The name of the file (.vhn) to be saved to or an instance of ostream derived class.
      void SaveVHN(const CDataSink &sink) const;
      //!
      //! Load from VortexGE's native VHN format.
      //! \param[in] source The name of the file (.vhn) to be loaded or an instance of istream derived class.
      void LoadVHN(const CDataSource &source);

      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

    protected:
      // I/O functions
      void _WriteNodeData_(ostream &os, const CHNode *pcsHNode) const;
      void _ReadNodeData_(pistream &is, CHNode *pcsHNode);

      // Data
      struct _TMatrixData {
        float          afScl[3];   // Dynamic scalling factor XYZ
        int32_t        an32Rot[3]; // Dynamic rotation factor XYZ
        float          afPos[3];   // Dynamic position factor XYZ
        CMatrix3D _sse csStc;      // Static transformation matrix
        CMatrix3D _sse csTot;      // Total transformation matrix
      } _sse _tM;
      struct _TMatrixCache {
        bool           bStc;     // Indicates that the static transformation has changed
        bool           bDyn;     // Indicates that the dynamic transformation has changed
        bool           bPsc;     // Indicates that the dynamic position has changed
        float          afPsc[3]; // Position change
        CMatrix3D _sse cs1;      // 1st cache
        CMatrix3D _sse cs2;      // 2nd cache
      } _sse _tC;
      CHNode  *_pcsParent;              // Pointer to parent node
      CHNode  *_apcsChild[_MAX_CHILD_]; // Pointer to child nodes
      uint32_t _un32UserID;             // User ID of this node
  };

  void CHNode::StcIdentity()
  {
    _tM.csStc.SetToIdentityMatrix();
    _tC.bStc = true;
  }

  void CHNode::StcShear(float fX, float fY)
  {
    _tM.csStc.Shear(fX, fY);
    _tC.bStc = true;
  }

  void CHNode::StcScale(float fX, float fY, float fZ)
  {
    _tM.csStc.Scale(fX, fY, fZ);
    _tC.bStc = true;
  }

  void CHNode::StcRotate(int32_t n32X, int32_t n32Y, int32_t n32Z)
  {
    _tM.csStc.Rotate(n32X, n32Y, n32Z);
    _tC.bStc = true;
  }

  void CHNode::StcTranslate(float fX, float fY, float fZ)
  {
    _tM.csStc.Translate(fX, fY, fZ);
    _tC.bStc = true;
  }

  void CHNode::StcSet(float const aafNewMatrix[4][4])
  {
    _tM.csStc.SetFromMatrix(aafNewMatrix);
    _tC.bStc = true;
  }

  void CHNode::StcSet(CMatrix3D &csVNewMatrix)
  {
    _tM.csStc.SetFromMatrix(csVNewMatrix);
    _tC.bStc = true;
  }

  void CHNode::DynSetIdentity()
  {
    _tM.afScl[0]   = 1; _tM.afScl[1]   = 1; _tM.afScl[2]   = 1;
    _tM.an32Rot[0] = 0; _tM.an32Rot[1] = 0; _tM.an32Rot[2] = 0;
    _tM.afPos[0]   = 0; _tM.afPos[1]   = 0; _tM.afPos[2]   = 0;
    _tC.bDyn = true;
  }

  void CHNode::DynSetScalling(float fX, float fY, float fZ)
  {
    _tM.afScl[0] = fX; _tM.afScl[1] = fY; _tM.afScl[2] = fZ;
    _tC.bDyn = true;
  }

  void CHNode::DynSetRotation(int32_t n32X, int32_t n32Y, int32_t n32Z)
  {
    _tM.an32Rot[0] = n32X; _tM.an32Rot[1] = n32Y; _tM.an32Rot[2] = n32Z;
    _tC.bDyn = true;
  }

  void CHNode::DynSetPosition(float fX, float fY, float fZ)
  {
    _tM.afPos[0] = fX; _tM.afPos[1] = fY; _tM.afPos[2] = fZ;
    _tC.bDyn = true;
  }

  void CHNode::DynScale(float fX, float fY, float fZ)
  {
    _tM.afScl[0] *= fX; _tM.afScl[1] *= fY; _tM.afScl[2] *= fZ;
    _tC.bDyn = true;
  }

  void CHNode::DynRotate(int32_t n32X, int32_t n32Y, int32_t n32Z)
  {
    _tM.an32Rot[0] += n32X; _tM.an32Rot[1] += n32Y; _tM.an32Rot[2] += n32Z;
    _tC.bDyn = true;
  }

  void CHNode::DynRotateX(int32_t n32X)
  {
    _tM.an32Rot[0] += n32X;
    _tC.bDyn = true;
  }

  void CHNode::DynRotateY(int32_t n32Y)
  {
    _tM.an32Rot[1] += n32Y;
    _tC.bDyn = true;
  }

  void CHNode::DynRotateZ(int32_t n32Z)
  {
    _tM.an32Rot[2] += n32Z;
    _tC.bDyn = true;
  }

  void CHNode::DynTranslate(float fX, float fY, float fZ)
  {
    _tM.afPos[0] += fX; _tM.afPos[1] += fY; _tM.afPos[2] += fZ;
    _tC.afPsc[0]  = fX; _tC.afPsc[1]  = fY; _tC.afPsc[2]  = fZ;
    _tC.bPsc = true;
  }

} // namespace vge
#endif
