// ############################################################################################
//  UnfoggedCeilingLineSegment.cc
//  Impelementation of unfogged ceiling-line-segment classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "FixedPointConsts.h"
#include "UnfoggedCeilingLineSegment.h"


namespace vge {

//
// The untextured-flat-shaded ceiling-line-segment class
//
_CUFCLineSeg::_CUFCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), ___dummy1(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), ___dummy2(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0),
  _n32XStep(0),               _n32ZStep(0),               _n32EdgeHeight(0)
#endif
{
}

void _CUFCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
  }
}

void _CUFCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}


//
// The untextured-gouraud-shaded ceiling-line-segment class
//
_CUGCLineSeg::_CUGCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), ___dummy1(1),
  _n32R    (0), _n32G    (0), _n32B    (0), ___dummy2(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), ___dummy3(1),
  _n32RStep(0), _n32GStep(0), _n32BStep(0), ___dummy4(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0),
  _n32R    (0), _n32G    (0), _n32B    (0),
  _n32XStep(0),               _n32ZStep(0),
  _n32RStep(0), _n32GStep(0), _n32BStep(0),               _n32EdgeHeight(0)
#endif
{
}

void _CUGCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32R = tP1.n32R;
  _n32G = tP1.n32G;
  _n32B = tP1.n32B;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32RStep = (tP2.n32R - tP1.n32R) / _n32EdgeHeight;
    _n32GStep = (tP2.n32G - tP1.n32G) / _n32EdgeHeight;
    _n32BStep = (tP2.n32B - tP1.n32B) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32RStep = 0;
    _n32GStep = 0;
    _n32BStep = 0;
  }
}

void _CUGCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32R += (_n32RStep * n32Step);
    _n32G += (_n32GStep * n32Step);
    _n32B += (_n32BStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}


//
// The textured-flat-shaded ceiling-line-segment class
//
_CTFCLineSeg::_CTFCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), ___dummy1(1),
  _n32U    (0), _n32V    (0), ___dummy2(1), ___dummy3(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), ___dummy4(1),
  _n32UStep(0), _n32VStep(0), ___dummy5(1), ___dummy6(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0),
  _n32U    (0), _n32V    (0),
  _n32XStep(0),               _n32ZStep(0),
  _n32UStep(0), _n32VStep(0),                             _n32EdgeHeight(0)
#endif
{
}

void _CTFCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32U = tP1.n32U;
  _n32V = tP1.n32V;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32UStep = (tP2.n32U - tP1.n32U) / _n32EdgeHeight;
    _n32VStep = (tP2.n32V - tP1.n32V) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32UStep = 0;
    _n32VStep = 0;
  }
}

void _CTFCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32U += (_n32UStep * n32Step);
    _n32V += (_n32VStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}


//
// The textured-gouraud-shaded ceiling-line-segment class
//
_CTGCLineSeg::_CTGCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), ___dummy1(1),
  _n32U    (0), _n32V    (0), ___dummy2(1), ___dummy3(1),
  _n32R    (0), _n32G    (0), _n32B    (0), ___dummy4(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), ___dummy5(1),
  _n32UStep(0), _n32VStep(0), ___dummy6(1), ___dummy7(1),
  _n32RStep(0), _n32GStep(0), _n32BStep(0), ___dummy8(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0),
  _n32U    (0), _n32V    (0),
  _n32R    (0), _n32G    (0), _n32B    (0),
  _n32XStep(0),               _n32ZStep(0),
  _n32UStep(0), _n32VStep(0),
  _n32RStep(0), _n32GStep(0), _n32BStep(0),               _n32EdgeHeight(0)
#endif
{
}

void _CTGCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32U = tP1.n32U;
  _n32V = tP1.n32V;
  _n32R = tP1.n32R;
  _n32G = tP1.n32G;
  _n32B = tP1.n32B;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32UStep = (tP2.n32U - tP1.n32U) / _n32EdgeHeight;
    _n32VStep = (tP2.n32V - tP1.n32V) / _n32EdgeHeight;
    _n32RStep = (tP2.n32R - tP1.n32R) / _n32EdgeHeight;
    _n32GStep = (tP2.n32G - tP1.n32G) / _n32EdgeHeight;
    _n32BStep = (tP2.n32B - tP1.n32B) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32UStep = 0;
    _n32VStep = 0;
    _n32RStep = 0;
    _n32GStep = 0;
    _n32BStep = 0;
  }
}

void _CTGCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32U += (_n32UStep * n32Step);
    _n32V += (_n32VStep * n32Step);
    _n32R += (_n32RStep * n32Step);
    _n32G += (_n32GStep * n32Step);
    _n32B += (_n32BStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}

} // namespace vge
