/** \file UString.h */
// ############################################################################################
//  UString.h
//  Definition & implementation of CUString class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___USTRING_H_
#define _BASE___USTRING_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {

  //!
  //! \brief Unicode string class
  //!
  //! For now it just a simple wrapper class, but later it
  //! may become a standalone unicode string class.
  //!
  //! \api beta
  class VGE_EXPORT CUString {
    public:
      enum ESCE { ESCE_Ascii, ESCE_Utf8 };

      typedef uint8_t  Utf8;
      typedef uint32_t Utf32;

      typedef basic_string<Utf32> Utf32String;

    public:
      //!
      //! Construct an empty CUString.
      inline CUString()
      : _str(1, 0)      // If not initialized with one NULL character first, it may
      { _str.clear(); } // cause segmentation fault. Bug from libstdc++ ?
      //!
      //! Construct a CUString from another CUString.
      inline CUString(const CUString &str)
      : _str(str._str)
      {}
      //!
      //! Construct a CUString from a Utf32String.
      inline CUString(const Utf32String &str)
      : _str(str)
      {}
      //!
      //! Construct a CUString from an std::string encoded in ASCII or UTF-8.
      inline CUString(const string &str, ESCE encoding = ESCE_Utf8)
      : _str(1, 0)
      {
        if(encoding == ESCE_Ascii) AsciiToUtf32(_str, str);
        else                       Utf8ToUtf32 (_str, str);
      }

      // Convenience functions
      //!
      //! Returns reference to the Utf32String contained in this CUString.
      inline Utf32String &utf32str()
      { return(_str); }
      //!
      //! Returns reference to the Utf32String contained in this CUString.
      inline const Utf32String &utf32str() const
      { return(_str); }

      // Informations functions
      //!
      //! Returns true if the string is empty.
      inline bool empty() const
      { return(_str.empty()); }
      //!
      //! Returns the length of the string.
      inline Utf32String::size_type size() const
      { return(_str.size()); }
      //!
      //! Returns the length of the string.
      inline Utf32String::size_type length() const
      { return(_str.length()); }

      // Direct data access functions
      //!
      //! Returns constant pointer to the string contents.
      const Utf32String::value_type *data () const
      { return(_str.data()); }
      //!
      //! Subscript access to the data contained in the string.
      inline Utf32String::reference operator[](Utf32String::size_type pos)
      { return(_str[pos]); }
      //!
      //! Subscript access to the data contained in the string.
      inline Utf32String::const_reference operator[](Utf32String::size_type pos) const
      { return(_str[pos]); }

      // Assigment operators
      //!
      //! Assigment operator from another CUString.
      inline CUString &operator = (const CUString &str)
      { _str = str._str; return(*this); }
      //!
      //! Assigment operator from a Utf32String.
      inline CUString &operator = (const Utf32String &str)
      { _str = str; return(*this); }
      //!
      //! Assigment operator from a C-string encoded in UTF-32.
      inline CUString &operator = (const Utf32String::value_type *str)
      { _str = str; return(*this); }

      // Appending operators
      //!
      //! Append a CUString to this string.
      inline CUString &operator += (const CUString &str)
      { _str += str._str; return(*this); }
      //!
      //! Append a Utf32String to this string.
      inline CUString &operator += (const Utf32String &str)
      { _str += str; return(*this); }
      //!
      //! Append a C-string encoded in UTF-32 to this string.
      inline CUString &operator += (const Utf32String::value_type *str)
      { _str += str; return(*this); }
      //!
      //! Append a UTF-32 character to this string.
      inline CUString &operator += (Utf32String::value_type c)
      { _str += c; return(*this); }

      // Insertion functions
      //!
      //! Insert a UTF-32 character at the end of this string
      inline void push_back(Utf32String::value_type c)
      { _str.push_back(c); }
      //!
      //! Insert a CUString.
      inline CUString &insert(Utf32String::size_type pos, const CUString &str)
      { _str.insert(pos, str._str); return(*this); }
      //!
      //! Insert a Utf32String.
      inline CUString &insert(Utf32String::size_type pos, const Utf32String &str)
      { _str.insert(pos, str); return(*this); }
      //!
      //! Insert multiple UTF-32 characters .
      inline CUString &insert(Utf32String::size_type pos, Utf32String::size_type n, Utf32String::value_type c)
      { _str.insert(pos, n, c); return(*this); }
      //!
      //! Insert a C-string encoded in ASCII or UTF-8.
      inline CUString &insert(Utf32String::size_type pos, const char *str, ESCE encoding = ESCE_Utf8)
      { CUString tmp(str, encoding); return(insert(pos, tmp)); }
      //!
      //! Insert a part of a CUString.
      inline CUString &insert(Utf32String::size_type pos, const CUString &str, Utf32String::size_type start, Utf32String::size_type n)
      { _str.insert(pos, str._str, start, n); return(*this); }
      //!
      //! Insert a part of a Utf32String.
      inline CUString &insert(Utf32String::size_type pos, const Utf32String &str, Utf32String::size_type start, Utf32String::size_type n)
      { _str.insert(pos, str, start, n); return(*this); }
      //!
      //! Insert a C-substring encoded in ASCII or UTF-8.
      inline CUString &insert(Utf32String::size_type pos, const char *str, Utf32String::size_type n, ESCE encoding = ESCE_Utf8)
      { CUString tmp(string(str, n), encoding); return(insert(pos, tmp)); }

      // Removing functions
      //!
      //! Remove some characters from this string.
      inline CUString &erase(Utf32String::size_type pos = 0, Utf32String::size_type n = Utf32String::npos)
      { _str.erase(pos, n); return(*this); }
      //!
      //! Erases the string (making it empty).
      inline void clear()
      { _str.clear(); }

      // Finding functions
      //!
      //! Find position of a CUString.
      Utf32String::size_type find(const CUString &str, Utf32String::size_type pos = 0) const
      { return(_str.find(str._str, pos)); }
      //!
      //! Find position of a Utf32String.
      Utf32String::size_type find(const Utf32String &str, Utf32String::size_type pos = 0) const
      { return(_str.find(str, pos)); }
      //!
      //! Find position of a UTF-32 character in this string.
      Utf32String::size_type find(Utf32String::value_type c, Utf32String::size_type pos = 0) const
      { return(_str.find(c, pos)); }
      //!
      //! Find last position of a CUString.
      Utf32String::size_type rfind(const CUString &str, Utf32String::size_type pos = Utf32String::npos) const
      { return(_str.rfind(str._str, pos)); }
      //!
      //! Find last position of a Utf32String.
      Utf32String::size_type rfind(const Utf32String &str, Utf32String::size_type pos = Utf32String::npos) const
      { return(_str.rfind(str, pos)); }
      //!
      //! Find last position of a UTF-32 character.
      Utf32String::size_type rfind(Utf32String::value_type c, Utf32String::size_type pos = Utf32String::npos) const
      { return(_str.rfind(c, pos)); }

      // Other string functions
      //!
      //! Resizes the string to the specified number of characters.
      void resize(Utf32String::size_type n, Utf32String::value_type c = 0)
      { _str.resize(n, c); }
      //!
      //! Get a substring.
      CUString substr(Utf32String::size_type pos = 0, Utf32String::size_type n = Utf32String::npos) const
      { return(CUString(_str.substr(pos, n))); }
      //!
      //! Swap contents with another CUString.
      void swap(CUString &s)
      { _str.swap(s._str); }

      // Conversion functions
      //!
      //! Convert this CUString to an std::string encoded in UTF-8.
      inline void ToUtf8(string &dst) const
      { Utf32ToUtf8(dst, &_str[0], _str.length()); }
      //!
      //! Convert this CUString to an std::string encoded in UTF-8.
      inline const string ToUtf8() const
      { string tmp; ToUtf8(tmp); return(tmp); }
      //!
      //! Perform case folding (needed for case-insensitive comparison).
      inline const CUString FoldCase(bool fullCaseFolding = true, bool specialCaseFolding = false) const
      { CUString tmp; CaseFolding(tmp, *this, fullCaseFolding, specialCaseFolding); return(tmp); }

      //!
      //! Convert a C-string encoded in ASCII to a Utf32String.
      static void AsciiToUtf32(Utf32String &dst, const char *src, size_t slen);
      //!
      //! Convert a C-string encoded in UTF-8 to a Utf32String.
      static void Utf8ToUtf32(Utf32String &dst, const Utf8 *src, size_t slen);
      //!
      //! Convert a C-string encoded in UTF-32 to an std::string encoded in UTF-8.
      static void Utf32ToUtf8(string &dst, const Utf32 *src, size_t slen);

      //!
      //! Convert an std::string encoded in ASCII to a Utf32String.
      inline static void AsciiToUtf32(Utf32String &dst, const string &src)
      { AsciiToUtf32(dst, src.data(), src.length()); }
      //!
      //! Convert an std::string encoded in UTF-8 to a Utf32String.
      inline static void Utf8ToUtf32(Utf32String &dst, const string &src)
      { Utf8ToUtf32(dst, reinterpret_cast<const Utf8*>(src.data()), src.length()); }
      //!
      //! Convert a Utf32String to an std::string encoded in UTF-8.
      inline static void Utf32ToUtf8(string &dst, const Utf32String &src)
      { Utf32ToUtf8(dst, src.data(), src.length()); }

      //!
      //! Perform case folding (needed for case-insensitive comparison).
      static void CaseFolding(CUString &dst, const CUString &src, bool fullCaseFolding = true, bool specialCaseFolding = false);

    protected:
      Utf32String _str;
  };


  //!
  //! Concatenate two CUStrings
  inline CUString operator + (const CUString &str1, const CUString &str2)
  {
    CUString strc(str1);
    strc += str2;
    return strc;
  }

} // namespace vge


namespace std {

  //!
  //! Swap the contents of two vge::CUString instances.
  inline void swap(vge::CUString &s1, vge::CUString &s2)
  { s1.swap(s2); }

}

#endif


/*
  ###############################################################################
  # This CUString class uses some of the source code which came from the files: #
  #   * ConvertUTF.h                                                            #
  #   * ConvertUTF.c                                                            #
  # which were taken from:                                                      #
  #   ftp://ftp.unicode.org/Public/PROGRAMS/CVTUTF                              #
  ###############################################################################
  # The original copyright notice is attached below                             #
  ###############################################################################
*/

/*
 * Copyright 2001-2004 Unicode, Inc.
 *
 * Disclaimer
 *
 * This source code is provided as is by Unicode, Inc. No claims are
 * made as to fitness for any particular purpose. No warranties of any
 * kind are expressed or implied. The recipient agrees to determine
 * applicability of information provided. If this file has been
 * purchased on magnetic or optical media from Unicode, Inc., the
 * sole remedy for any claim will be exchange of defective media
 * within 90 days of receipt.
 *
 * Limitations on Rights to Redistribute This Code
 *
 * Unicode, Inc. hereby grants the right to freely use the information
 * supplied in this file in the creation of products supporting the
 * Unicode Standard, and to make copies of this file in any form
 * for internal or external distribution as long as this notice
 * remains attached.
 */
