// ############################################################################################
//  Plane3DBuilder.cc
//  Implementation of CPlane3DBuilder and CPlane3D classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "Plane3DBuilder.h"

// C++ headers
#include <sstream>


namespace vge {

//
// Constants used in all execution CPlane3DBuilder data file (*.vpb)
//
static const uint32_t  _VPB_FILE_VER_ = 100;
static const char     *_VPB_MAGIC_    = "VPB";


//
// CPlane3D implementation
//
const TFloatXYZ CPlane3D::DEFAULT_GRID_SIZE(100, 100, 100);

void CPlane3D::TranslateVertices(const TFloatXYZ &factor)
{
  const float &fx = factor.x;
  const float &fy = factor.y;
  const float &fz = factor.z;

  // Translate vertices
  size_t n = 0;
  for(size_t i = 0; i < _vertices.size()/3; ++i) {
    _vertices[n++] += fx;
    _vertices[n++] += fy;
    _vertices[n++] += fz;
  }

  // Translate the plane origin
  _planeOrigin.x += fx;
  _planeOrigin.y += fy;
  _planeOrigin.z += fz;
}

void CPlane3D::RandomlyVaryHeights(float              maxFactor,
                                   bool               inProcentageOfGridSize,
                                   bool               excludeOutlineVertices,
                                   CRandomCoreUInt32 *randomCore)
{
  // Calculate the movement based on the plane's direction
  float xf = 0, yf = 0, zf = 0;
  switch(_planeDir) {
    case EPD_0P0: yf =  1; break; // as if it is the top side of a 3D cube
    case EPD_0N0: yf = -1; break; // as if it is the bottom side of a 3D cube
    case EPD_P00: xf =  1; break; // as if it is the right side of a 3D cube
    case EPD_N00: xf = -1; break; // as if it is the left side of a 3D cube
    case EPD_00P: zf =  1; break; // as if it is the back side of a 3D cube
    case EPD_00N: zf = -1; break; // as if it is the front side of a 3D cube
  }
  if(inProcentageOfGridSize) {
    const float ff = maxFactor / 100.0f;
    xf *= (_gridSize.x * ff);
    yf *= (_gridSize.y * ff);
    zf *= (_gridSize.z * ff);
  }
  else {
    xf *= maxFactor;
    yf *= maxFactor;
    zf *= maxFactor;
  }

  // Used for generating random number in the range of 0.0 to 1.0
  CRandomCoreUInt32 defaultCore;
  CRandomFloat      randomFloat(randomCore ? *randomCore : defaultCore);

  // Walk trough all the vertices and randomly vary the plane's height,
  // except for the outline (border) vertices
  if(excludeOutlineVertices) {
    // Calculate the lower and upper limit
    const uint32_t vc = _planeColSize + 1;
    const uint32_t ll = _planeColSize;
    const uint32_t ul = _planeRowSize * vc;
    // Modify the vertices
    size_t n = 0;
    for(size_t i = 0; i < _vertices.size()/3; ++i) {
      // Check if it is a party of the outline (border) vertices
      const uint32_t r = i % vc;
      if(i<=ll || i>=ul || r==0 || r==_planeColSize) {
        n += 3;
        continue;
      }
      // Vary its height
      const float rf = randomFloat.GeneratePositive();
      _vertices[n++] += (xf * rf);
      _vertices[n++] += (yf * rf);
      _vertices[n++] += (zf * rf);
    }
  }
  // Walk trough all the vertices and randomly vary the plane's height
  else {
    size_t n = 0;
    for(size_t i = 0; i < _vertices.size()/3; ++i) {
      const float rf = randomFloat.GenerateBipolar();
      _vertices[n++] += (xf * rf);
      _vertices[n++] += (yf * rf);
      _vertices[n++] += (zf * rf);
    }
  }
}

void CPlane3D::GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const
{
  minCoord.x = minCoord.y = minCoord.z =  numeric_limits<float>::max();
  maxCoord.x = maxCoord.y = maxCoord.z = -numeric_limits<float>::max();

  // Walk trough all the vertices and find their minimum and maximum coordinates
  size_t n = 0;
  for(size_t i = 0; i < _vertices.size()/3; ++i) {
    const float x = _vertices[n++];
    const float y = _vertices[n++];
    const float z = _vertices[n++];

    if(x < minCoord.x) minCoord.x = x; if(x > maxCoord.x) maxCoord.x = x;
    if(y < minCoord.y) minCoord.y = y; if(y > maxCoord.y) maxCoord.y = y;
    if(z < minCoord.z) minCoord.z = z; if(z > maxCoord.z) maxCoord.z = z;
  }
}

void CPlane3D::GetOriginPosition(TFloatXYZ &orgCoord) const
{
  TFloatXYZ mn, mx;
  GetBoundingBox(mn, mx);
  switch(_planeDir) {
    case EPD_0P0: orgCoord = TFloatXYZ(mn.x, mn.y, mx.z); break; // as if it is the top side of a 3D cube
    case EPD_0N0: orgCoord = TFloatXYZ(mn.x, mn.y, mn.z); break; // as if it is the bottom side of a 3D cube
    case EPD_P00: orgCoord = TFloatXYZ(mn.x, mx.y, mn.z); break; // as if it is the right side of a 3D cube
    case EPD_N00: orgCoord = TFloatXYZ(mn.x, mx.y, mx.z); break; // as if it is the left side of a 3D cube
    case EPD_00P: orgCoord = TFloatXYZ(mx.x, mx.y, mn.z); break; // as if it is the back side of a 3D cube
    case EPD_00N: orgCoord = TFloatXYZ(mn.x, mx.y, mn.z); break; // as if it is the front side of a 3D cube
  }
}

CPlane3D &CPlane3D::operator=(const CPlane3D &src)
{
  _vertices     = src._vertices;
  _faces        = src._faces;
  _gridSize     = src._gridSize;
  _planeOrigin  = src._planeOrigin;
  _planeRowSize = src._planeRowSize;
  _planeColSize = src._planeColSize;
  _planeDir     = src._planeDir;
  return(*this);
}

/*
  -------------------------------------
  _GeneratePlane() Implementation Notes
  -------------------------------------

  In the ascii-art below :
    * Xs are vertices
    * Rectangles are faces
    * Numbers without pharenthesis are vertex number
    * Numbers with pharenthesis are face number

     00       01       02       03       04       05
    X--------X--------X--------X--------X--------X
    |        |        |        |        |        |
    |  (00)  |  (01)  |  (02)  |  (03)  |  (04)  |
    |        |        |        |        |        |
    |06      |07      |08      |09      |10      |11
    X--------X--------X--------X--------X--------X
    |        |        |        |        |        |
    |  (05)  |  (06)  |  (07)  |  (08)  |  (09)  |
    |        |        |        |        |        |
    |12      |13      |14      |15      |16      |17
    X--------X--------X--------X--------X--------X
    |        |        |        |        |        |
    |  (10)  |  (11)  |  (12)  |  (13)  |  (14)  |
    |        |        |        |        |        |
    |18      |19      |20      |21      |22      |23
    X--------X--------X--------X--------X--------X
    |        |        |        |        |        |
    |  (15)  |  (16)  |  (17)  |  (18)  |  (19)  |
    |        |        |        |        |        |
    |24      |25      |26      |27      |28      |29
    X--------X--------X--------X--------X--------X

  Rules of the game :
    * A plane is defined in row and columns, thus the above plane is
      a 4x5 plane.
    * The plane's origin is always at the top-left corner of the plane
      (the vertex #0)

  Equations :
    * Number of faces    -> nof = row x col
    * Number of vertices -> nov = (row + 1) x (col + 1)
    * At face #n :
        - Start of vertex row    -> svr = n div col
        - Start of vertex column -> svc = n % col
        - Start of vertex index  -> svi = svr x (col + 1) + svc
        - Indices                -> a = svi
                                    b = a + 1
                                    c = b + col
                                    d = c + 1
        - Texture map :
          a        b       0,0      1,0   (u,v)
          X--------X        X--------X
          |        |        |        |
          |  (#n)  |   =>   |        |
          |        |        |        |
          |        |        |        |
          X--------X        X--------X
          c        d       0,1      1,1
*/
void CPlane3D::_GeneratePlane(uint32_t planeRowSize, uint32_t planeColSize, EPlaneDir planeDir)
{
  const size_t  row  = planeRowSize; // Number of face rows
  const size_t  col  = planeColSize; // Number of face columns
  const size_t  row1 = row + 1;      // Number of vertex rows
  const size_t  col1 = col + 1;      // Number of vertex columns
  const size_t  nof  = row * col;    // Number of faces
  const size_t  nov  = row1 * col1;  // Number of vertices
  const float  &xgs  = _gridSize.x;  // X grid size
  const float  &ygs  = _gridSize.y;  // Y grid size
  const float  &zgs  = _gridSize.z;  // Z grid size

  // For clarity
  vector<float>    &verticesOut = _vertices;
  vector<uint32_t> &facesOut    = _faces;

  // Allocate memory
  facesOut.resize(nof * 4);
  verticesOut.resize(nov * 3);

  // Calculate the movement based on the plane's direction
  float dxr = 0, dyr = 0, dzr = 0;
  float dxc = 0, dyc = 0, dzc = 0;
  switch(planeDir) {
    case EPD_0P0: dzr = -zgs; dxc =  xgs; break; // as if it is the top side of a 3D cube
    case EPD_0N0: dzr =  zgs; dxc =  xgs; break; // as if it is the bottom side of a 3D cube
    case EPD_P00: dyr = -ygs; dzc =  zgs; break; // as if it is the right side of a 3D cube
    case EPD_N00: dyr = -ygs; dzc = -zgs; break; // as if it is the left side of a 3D cube
    case EPD_00P: dyr = -ygs; dxc = -xgs; break; // as if it is the back side of a 3D cube
    case EPD_00N: dyr = -ygs; dxc =  xgs; break; // as if it is the front side of a 3D cube
    default: VError(VERR_InvalidParameter, _NFL_, "Invalid plane direction");
  };

  // Generate the vertices
  size_t nv = 0;
  float  ix = 0;
  float  iy = 0;
  float  iz = 0;
  // Loop trough the rows
  for(size_t r = 0; r < row1; ++r) {
    // Calculate initial coordinate
    float jx = ix; ix += dxr;
    float jy = iy; iy += dyr;
    float jz = iz; iz += dzr;
    // Loop trough the columns
    for(size_t c = 0; c < col1; ++c) {
      // Put the vertex in the array
      verticesOut[nv++] = jx + _planeOrigin.x;
      verticesOut[nv++] = jy + _planeOrigin.y;
      verticesOut[nv++] = jz + _planeOrigin.z;
      // Update the coordinate
      jx += dxc;
      jy += dyc;
      jz += dzc;
    }
  }

  // Generate the faces
  size_t nf = 0;
  // Loop trough the faces
  for(size_t n = 0; n < nof; ++n) {
    // Calculate the vertices that belongs to this #n face
    const size_t svr = n / col;
    const size_t svc = n % col;
    const size_t svi = (svr * col1) + svc;
    const size_t a   = svi;
    const size_t b   = a + 1;
    const size_t c   = b + col;
    const size_t d   = c + 1;
    // Put the indices in the array
    facesOut[nf++] = c;
    facesOut[nf++] = a;
    facesOut[nf++] = b;
    facesOut[nf++] = d;
  }

  // Copy variables
  _planeRowSize = planeRowSize;
  _planeColSize = planeColSize;
  _planeDir     = planeDir; 
}

void CPlane3D::_ChangePlaneOrigin(const TFloatXYZ &newPlaneOrigin, bool updateExistingPlane)
{
  // Update existing vertices if asked
  if(updateExistingPlane && !_vertices.empty()) {
    // Calculate the movements
    const float dx = newPlaneOrigin.x - _planeOrigin.x;
    const float dy = newPlaneOrigin.y - _planeOrigin.y;
    const float dz = newPlaneOrigin.z - _planeOrigin.z;
    // Walk trough all the vertices and update their coordinates
    size_t n = 0;
    for(size_t i = 0; i < _vertices.size()/3; ++i) {
      _vertices[n++] += dx;
      _vertices[n++] += dy;
      _vertices[n++] += dz;
    }
  }

  // Copy the new plane origin
  _planeOrigin = newPlaneOrigin;
}

void CPlane3D::LoadFromStream(pistream &is)
{
  // Load the vertices
  vector<float> vs;
  uint32_t      sz;
  is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) {
    vs.resize(sz);
    is.read(reinterpret_cast<char*>(&vs[0]), sz * sizeof(float));
  }

  // Load the faces
  vector<uint32_t> fs;
  is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) {
    fs.resize(sz);
    is.read(reinterpret_cast<char*>(&fs[0]), sz * sizeof(uint32_t));
  }

  // Load the grid size
  TFloatXYZ gz;
  is.read(reinterpret_cast<char*>(&gz.x), sizeof(gz.x));
  is.read(reinterpret_cast<char*>(&gz.y), sizeof(gz.y));
  is.read(reinterpret_cast<char*>(&gz.z), sizeof(gz.z));

  // Load the plane origin
  TFloatXYZ po;
  is.read(reinterpret_cast<char*>(&po.x), sizeof(po.x));
  is.read(reinterpret_cast<char*>(&po.y), sizeof(po.y));
  is.read(reinterpret_cast<char*>(&po.z), sizeof(po.z));

  // Load plane size
  uint32_t rz, cz;
  is.read(reinterpret_cast<char*>(&rz), sizeof(rz));
  is.read(reinterpret_cast<char*>(&cz), sizeof(cz));

  // Load the plane direction
  uint32_t pd;
  is.read(reinterpret_cast<char*>(&pd), sizeof(pd));

  // Move variables
  swap(_vertices, vs);
  swap(_faces, fs);
  _gridSize     = gz;
  _planeOrigin  = po;
  _planeRowSize = rz;
  _planeColSize = cz;
  _planeDir     = static_cast<EPlaneDir>(pd);
}

void CPlane3D::SaveToStream(ostream &os) const
{
  // Save the vertices
  uint32_t sz = _vertices.size();
  os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) os.write(reinterpret_cast<const char*>(&_vertices[0]), sz * sizeof(float));

  // Save the faces
  sz = _faces.size();
  os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) os.write(reinterpret_cast<const char*>(&_faces[0]), sz * sizeof(uint32_t));

  // Save the grid size
  os.write(reinterpret_cast<const char*>(&_gridSize.x), sizeof(_gridSize.x));
  os.write(reinterpret_cast<const char*>(&_gridSize.y), sizeof(_gridSize.y));
  os.write(reinterpret_cast<const char*>(&_gridSize.z), sizeof(_gridSize.z));

  // Save the plane origin
  os.write(reinterpret_cast<const char*>(&_planeOrigin.x), sizeof(_planeOrigin.x));
  os.write(reinterpret_cast<const char*>(&_planeOrigin.y), sizeof(_planeOrigin.y));
  os.write(reinterpret_cast<const char*>(&_planeOrigin.z), sizeof(_planeOrigin.z));

  // Save plane size
  os.write(reinterpret_cast<const char*>(&_planeRowSize), sizeof(_planeRowSize));
  os.write(reinterpret_cast<const char*>(&_planeColSize), sizeof(_planeColSize));

  // Save the plane direction
  uint32_t pd = _planeDir;
  os.write(reinterpret_cast<char*>(&pd), sizeof(pd));
}


//
// CPlane3DBuilder implementation
//
const TFloatXYZ &CPlane3DBuilder::DEFAULT_GRID_SIZE   = CPlane3D::DEFAULT_GRID_SIZE;
const float      CPlane3DBuilder::DEFAULT_ERROR_LIMIT = 0.001f;

CPlane3DBuilder::~CPlane3DBuilder() throw()
{
  ClearAll();
}

void CPlane3DBuilder::ClearAll()
{
  for(CPlane3DMap::iterator it = _planes.begin(); it != _planes.end(); ++it)
    delete it->second;

  _planes.clear();
}

void CPlane3DBuilder::GeneratePlane(const string    &planeName,
                                    const TFloatXYZ &planeOrigin,
                                    uint32_t         planeRowSize,
                                    uint32_t         planeColSize,
                                    EPlaneDir        planeDir)
{
  // Check if already exist
  if(_planes.find(planeName) != _planes.end())
    VError(VERR_Exist, _NFL_, "A CPlane3D instance with the same name (\"%s\") is already exist", planeName.c_str());

  // Create a new CPlane3D instance
  CPlane3D *p3d = new CPlane3D(_gridSize, planeOrigin);

  // Create the plane
  p3d->GeneratePlane(planeRowSize, planeColSize, planeDir);

  // Put it in the map
  _planes[planeName] = p3d;
}

void CPlane3DBuilder::RenamePlane(const string &oldPlaneName, const string &newPlaneName)
{
  // Check if not exist
  CPlane3DMap::iterator it = _planes.find(oldPlaneName);
  if(it == _planes.end())
    VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\"", oldPlaneName.c_str());

  // Check if already exist
  if(_planes.find(newPlaneName) != _planes.end())
    VError(VERR_Exist, _NFL_, "A CPlane3D instance with the same name (\"%s\") is already exist", newPlaneName.c_str());

  // Get the instance, delete the old map entry and put the new entry in the map
  CPlane3D *p3d = it->second;
  _planes.erase(it);
  _planes[newPlaneName] = p3d ;
}

void CPlane3DBuilder::DeletePlane(const string &planeName)
{
  CPlane3DMap::iterator it = _planes.find(planeName);
  if(it == _planes.end()) return;

  delete it->second;
  _planes.erase(it);
}

void CPlane3DBuilder::DeletePlane(const vector<string> &planeNames)
{
  if(_planes.empty()) return;
  for(size_t i = 0; i < planeNames.size(); ++i)
    DeletePlane(planeNames[i]);
}

void CPlane3DBuilder::TranslateVertices(const TFloatXYZ &factor)
{
  for(CPlane3DMap::iterator it = _planes.begin(); it != _planes.end(); ++it) {
    it->second->TranslateVertices(factor);
  }
}

void CPlane3DBuilder::TranslateVertices(const TFloatXYZ      &factor,
                                        const vector<string> &selectedPlanesName,
                                        bool                  ignoreIfNotExist)
{
  for(size_t n = 0; n < selectedPlanesName.size(); ++n) {
    // Get the source plane
    CPlane3DMap::iterator it = _planes.find(selectedPlanesName[n]);
    if(it == _planes.end()) {
      if(ignoreIfNotExist) continue;
      VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\"", selectedPlanesName[n].c_str());
    }
    // Translate its vertices
    it->second->TranslateVertices(factor);
  }
}

void CPlane3DBuilder::Combine(const CPlane3DBuilder &src, const string &namePrefix)
{
  // Prepare the name prefix
  const string np = namePrefix.empty() ? "" : (namePrefix + '-');

  // Copy all the planes
  for(CPlane3DMap::const_iterator it = src._planes.begin(); it != src._planes.end(); ++it) {
    // Prepare the name
    const string oname = np + it->first;
          string uname = oname;
    for(size_t i = 0; ; ++i) {
      // Break if the name is unique
      if(_planes.find(uname) == _planes.end()) break;
      // Adjust the name
      stringstream ss;
      ss << i;
      uname = oname + '-' + ss.str();
    }
    // Copy the plane
    CPlane3D *p3d = new CPlane3D(*it->second);
    // Put it in the map
    _planes[uname] = p3d;
  }
}

void CPlane3DBuilder::Combine(const CPlane3DBuilder &src,
                              const vector<string>  &srcPlaneNames,
                              const string          &namePrefix,
                              bool                  ignoreIfNotExist)
{
  // Prepare the name prefix
  const string np = namePrefix.empty() ? "" : (namePrefix + '-');

  // Copy all the planes
  for(size_t n = 0; n < srcPlaneNames.size(); ++n) {
    // Get the source plane
    CPlane3DMap::const_iterator it = src._planes.find(srcPlaneNames[n]);
    if(it == src._planes.end()) {
      if(ignoreIfNotExist) continue;
      VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\" in the source CPlane3DBuilder instance", srcPlaneNames[n].c_str());
    }
    // Prepare the name
    const string oname = np + it->first;
          string uname = oname;
    for(size_t i = 0; ; ++i) {
      // Break if the name is unique
      if(_planes.find(uname) == _planes.end()) break;
      // Adjust the name
      stringstream ss;
      ss << i;
      uname = oname + '-' + ss.str();
    }
    // Copy the plane
    CPlane3D *p3d = new CPlane3D(*it->second);
    // Put it in the map
    _planes[uname] = p3d;
  }
}

void CPlane3DBuilder::GeneratePanel3DBuffer(CPanel3DBuffer &dstPanel3DBuffer, float combineVertexErrorLimit) const
{
  // Combine the vertices and faces
  vector<float>    cv;
  vector<uint32_t> cf;
  for(CPlane3DMap::const_iterator it = _planes.begin(); it != _planes.end(); ++it) {
    _CombinePlanes(cv, cf, it->second->GetVertices(), it->second->GetFaces(), combineVertexErrorLimit);
  }
  //const size_t cz = cv.size() / 3;
  //for(size_t i = 0; i < cf.size(); ++i) {
  //  vassert(cf[i] < cz, "Indices #%d has an invalid value", i);
  //}

  // Convert to CPanel3DBuffer
  dstPanel3DBuffer.InitFromMemory(&cv[0], cv.size()/3, &cf[0], cf.size()/4);
}

void CPlane3DBuilder::GeneratePanel3DBuffer(CPanel3DBuffer &dstPanel3DBuffer,
                                            CTexture       *textureToBeUsed,
                                            float           combineVertexErrorLimit) const
{
  GeneratePanel3DBuffer(dstPanel3DBuffer, combineVertexErrorLimit);
  dstPanel3DBuffer.SetTexture(textureToBeUsed);
  dstPanel3DBuffer.CalcDefTexCoords();
}

void CPlane3DBuilder::GeneratePanel3DBuffer(CPanel3DBuffer       &dstPanel3DBuffer,
                                            const vector<string> &selectedPlanesName,
                                            float                 combineVertexErrorLimit,
                                            bool                  ignoreIfNotExist) const
{
  // Combine the vertices and faces
  vector<float>    cv;
  vector<uint32_t> cf;
  
  // Copy all the planes
  for(size_t n = 0; n < selectedPlanesName.size(); ++n) {
    // Get the plane
    CPlane3DMap::const_iterator it = _planes.find(selectedPlanesName[n]);
    if(it == _planes.end()) {
      if(ignoreIfNotExist) continue;
      VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\"", selectedPlanesName[n].c_str());
     }
    // Combine it
    _CombinePlanes(cv, cf, it->second->GetVertices(), it->second->GetFaces(), combineVertexErrorLimit);
  }

  // Convert to CPanel3DBuffer
  dstPanel3DBuffer.InitFromMemory(&cv[0], cv.size()/3, &cf[0], cf.size()/4);
}

void CPlane3DBuilder::GeneratePanel3DBuffer(CPanel3DBuffer       &dstPanel3DBuffer,
                                            const vector<string> &selectedPlanesName,
                                            CTexture             *textureToBeUsed,
                                            float                 combineVertexErrorLimit,
                                            bool                  ignoreIfNotExist) const
{
  GeneratePanel3DBuffer(dstPanel3DBuffer, selectedPlanesName, combineVertexErrorLimit, ignoreIfNotExist);
  dstPanel3DBuffer.SetTexture(textureToBeUsed);
  dstPanel3DBuffer.CalcDefTexCoords();
}

void CPlane3DBuilder::GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const
{
  minCoord.x = minCoord.y = minCoord.z =  numeric_limits<float>::max();
  maxCoord.x = maxCoord.y = maxCoord.z = -numeric_limits<float>::max();

  // Walk trough all the planes and find their minimum and maximum coordinates
  for(CPlane3DMap::const_iterator it = _planes.begin(); it != _planes.end(); ++it) {
    TFloatXYZ mn, mx;
    it->second->GetBoundingBox(mn, mx);

    if(mn.x < minCoord.x) minCoord.x = mn.x; if(mx.x > maxCoord.x) maxCoord.x = mx.x;
    if(mn.y < minCoord.y) minCoord.y = mn.y; if(mx.y > maxCoord.y) maxCoord.y = mx.y;
    if(mn.z < minCoord.z) minCoord.z = mn.z; if(mx.z > maxCoord.z) maxCoord.z = mx.z;
 }
}

void CPlane3DBuilder::GetPlaneNames(vector<string> dstBuff) const
{
  dstBuff.clear();
  for(CPlane3DMap::const_iterator it = _planes.begin(); it != _planes.end(); ++it) {
    dstBuff.push_back(it->first);
  }
}

CPlane3D &CPlane3DBuilder::operator [] (const string &planeName)
{
  CPlane3DMap::iterator it = _planes.find(planeName);
  _if_unlikely(it == _planes.end())
    VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\"", planeName.c_str());

  return(*it->second);
}

const CPlane3D &CPlane3DBuilder::operator [] (const string &planeName) const
{
  CPlane3DMap::const_iterator it = _planes.find(planeName);
  _if_unlikely(it == _planes.end())
    VError(VERR_NotExist, _NFL_, "There is no CPlane3D instance with the name \"%s\"", planeName.c_str());

  return(*it->second);
}

void CPlane3DBuilder::_CombinePlanes(vector<float>          &verticesOut,
                                     vector<uint32_t>       &facesOut,
                                     const vector<float>    &verticesIn,
                                     const vector<uint32_t> &facesIn,
                                     float                   errLimit) const
{
  // If the destination buffers are still empty, just copy directly
  // from the input buffers to the output buffers.
  if(verticesOut.empty() && facesOut.empty()) {
    verticesOut = verticesIn;
    facesOut    = facesIn;
    return;
  }

  // Get the number of vertices which are already inside the output
  // vertices buffer
  const size_t nov = verticesOut.size()/3;

  // For converting indices (vertices' indexes)
  map<uint32_t, uint32_t> indicesMap;

  // Copy the vertices from the input buffer to the output buffer
  size_t nv = 0;
  for(size_t i = 0; i < verticesIn.size()/3; ++i) {
    // Extract the new (input) vertex
    const float xNew = verticesIn[nv++];
    const float yNew = verticesIn[nv++];
    const float zNew = verticesIn[nv++];
    // Walk trough in the output vertices buffer and search for old
    // vertex which is very similar with new new vertex
    bool   dv = false;
    size_t ov = 0;
    for(size_t j = 0; j < nov; ++j) {
      // Extract the old (output) vertex
      const float xOld = verticesOut[ov++];
      const float yOld = verticesOut[ov++];
      const float zOld = verticesOut[ov++];
      // Calculate the cumulative absolute error
      const float xErr = fabs(xOld - xNew);
      const float yErr = fabs(yOld - yNew);
      const float zErr = fabs(zOld - zNew);
      const float cErr = xErr + yErr + zErr;
      // If the the cumulative absolute error is smaller than the error
      // limit, use the old vertex from the output vertices buffer
      if(cErr < errLimit) {
        //if(indicesMap.find(i) != indicesMap.end()) VError(VERR_Exist, _NFL_, "Duplicated vertex index in the indices convertion table");
        dv            = true;
        indicesMap[i] = j;
        //VPrint("SrcI %03d -> DstI %03d\n", i, j);
        break;
      }
    }
    // Put the new vertex only if there is no similar vertex in the
    // output vertices buffer
    if(!dv) {
      //if(indicesMap.find(i) != indicesMap.end()) VError(VERR_Exist, _NFL_, "Duplicated vertex index in the indices convertion table");
      indicesMap[i] = verticesOut.size()/3;
      verticesOut.push_back(xNew);
      verticesOut.push_back(yNew);
      verticesOut.push_back(zNew);
    }
  }

  // Copy the faces
  size_t nf = 0;
  for(size_t i = 0; i < facesIn.size()/4; ++i) {
    // Extract the new (input) indices
    uint32_t id[4];
    for(size_t j = 0; j < 4; ++j) {
      id[j] = facesIn[nf++];
      //vassert(indicesMap.find(id[j]) != indicesMap.end(),
      //        "Vertex index is not in the indices convertion table");
    }
    // Convert the extracted indices
    uint32_t newIndices[4];
    for(size_t j = 0; j < 4; ++j) newIndices[j] = indicesMap[id[j]];
    // Check if a face which is exactly the same with the new one already exist
    // in the destination face buffer
    bool   found = false;
    size_t of    = 0;
    for(size_t j = 0; j < facesOut.size()/4; ++j) {
      found = true;
      for(size_t k = 0; k < 4; ++k) found &= (newIndices[k] == facesOut[of++]);
      if(found) break;
    }
    if(found) continue;
    // Generate a new face
    for(size_t j = 0; j < 4; ++j) facesOut.push_back(newIndices[j]);
  }
}

void CPlane3DBuilder::SaveVPB(const CDataSink &sink) const
{
  // Preparation
  ostream    &os = sink.Stream();
  const char *nm = sink.Name().c_str();

  // Save the data
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    // Write the magic numbers
    os.write(_VPB_MAGIC_, strlen(_VPB_MAGIC_));
    // Write the version number
    uint32_t ver = _VPB_FILE_VER_;
    os.write(reinterpret_cast<char*>(&ver), sizeof(ver));
    // Write the grid size
    os.write(reinterpret_cast<const char*>(&_gridSize.x), sizeof(_gridSize.x));
    os.write(reinterpret_cast<const char*>(&_gridSize.y), sizeof(_gridSize.y));
    os.write(reinterpret_cast<const char*>(&_gridSize.z), sizeof(_gridSize.z));
    // Write the total number of planes
    uint32_t sz = _planes.size();
    os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
    // Write all the planes' data
    for(CPlane3DMap::const_iterator it = _planes.begin(); it != _planes.end(); ++it) {
      // Write the plane's name
      uint32_t nz = it->first.length();
      os.write(reinterpret_cast<char*>(&nz), sizeof(nz));
      if(nz) os.write(reinterpret_cast<const char*>(it->first.data()), nz);
      // Write the plane's data
      it->second->SaveToStream(os);
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write CPlane3DBuilder data to stream \"%s\"", nm);
  }
  catch(...) {
    throw;
  }
}

void CPlane3DBuilder::LoadVPB(const CDataSource &source)
{
  // Preparation
  pistream   &is = source.PStream();
  const char *nm = source.Name().c_str();

  // Variables for loading data
  TFloatXYZ   gz;
  CPlane3DMap pm;

  // Save the data
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    // Read the magic numbers
    char mnb[64];
    memset(mnb, 0, sizeof(mnb));
    is.read(mnb, strlen(_VPB_MAGIC_));
    if(strcmp(mnb, _VPB_MAGIC_)) VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data");
    // Read the version number
    uint32_t ver;
    is.read(reinterpret_cast<char*>(&ver), sizeof(ver));
    if(ver != _VPB_FILE_VER_) VError(VERR_IncorrectVersion, _NFL_, "Invalid version number");
    // Read the grid size
    is.read(reinterpret_cast<char*>(&gz.x), sizeof(gz.x));
    is.read(reinterpret_cast<char*>(&gz.y), sizeof(gz.y));
    is.read(reinterpret_cast<char*>(&gz.z), sizeof(gz.z));
    // Read the total number of planes
    uint32_t sz;
    is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
    // Read all the planes' data
    for(size_t i = 0; i < sz; ++i) {
      // Read the plane's name
      string   pn;
      uint32_t nz;
      is.read(reinterpret_cast<char*>(&nz), sizeof(nz));
      if(nz) {
        pn.resize(nz);
        is.read(reinterpret_cast<char*>(&pn[0]), nz);
      }
      // Create a new plane and put it in the map
      CPlane3D *p3d = new CPlane3D;
      try {
        pm[pn] = p3d;
      }
      catch(...) {
        delete p3d;
        throw;
      }
      // Read the plane's data
      p3d->LoadFromStream(is);
    }
  }
  catch(ios_base::failure&) {
    for(CPlane3DMap::const_iterator it = pm.begin(); it != pm.end(); ++it) {
      delete it->second;
    }
    VError(VERR_WriteStream, _NFL_, "Couldn't write CPlane3DBuilder data to stream \"%s\"", nm);
  }
  catch(...) {
    for(CPlane3DMap::const_iterator it = pm.begin(); it != pm.end(); ++it) {
      delete it->second;
    }
    throw;
  }

  // Move data
  ClearAll();
  swap(_planes, pm);
  _gridSize = gz;
}

void CPlane3DBuilder::CreateSimpleOpenRoom(const TFloatXYZ &roomOrigin,
                                           const TInt32XYZ &roomSize,
                                           int32_t         leftWallThickness,
                                           int32_t         rightWallThickness,
                                           int32_t         backWallThickness,
                                           int32_t         floorThickness,
                                           bool            fullWall,
                                           const string   &namePrefix,
                                           bool            deleteExistingPlanes)
{
  // Check the parameters
  if(roomSize.x<0 || roomSize.y<0 || roomSize.z<0)
    VError(VERR_InvalidParameter, _NFL_, "All the XYZ components of the room size must be greater than zero");

  // For convenience
  typedef TFloatXYZ _F;

  // Just for shortcut
  CPlane3DBuilder &p3db = *this;
  const float     &ox   = roomOrigin.x;       // X origin of the room
  const float     &oy   = roomOrigin.y;       // Y origin of the room
  const float     &oz   = roomOrigin.z;       // Z origin of the room
  const uint32_t  &sx   = roomSize.x;         // Size of the room in the X direction
  const uint32_t  &sy   = roomSize.y;         // Size of the room in the Y direction
  const uint32_t  &sz   = roomSize.z;         // Size of the room in the Z direction
  const int32_t   &lwt  = leftWallThickness;  // Thickness of the left wall
  const int32_t   &rwt  = rightWallThickness; // Thickness of the right wall
  const int32_t   &bwt  = backWallThickness;  // Thickness of the back wall
  const int32_t   &ft   = floorThickness;     // Thickness of the floot
  const bool      &fw   = fullWall;           // Should all the walls be generated ?

  // Prepare the name prefix
  const string np = namePrefix.empty() ? "" : (namePrefix + '-');

  // Preparation
  const uint32_t &wl = max<int32_t>(lwt, 0); // Thickness of the left wall
  const uint32_t &wr = max<int32_t>(rwt, 0); // Thickness of the right wall
  const uint32_t &wb = max<int32_t>(bwt, 0); // Thickness of the back wall
  const uint32_t  sb = sx + wl + wr;         // Size of the back wall
  const uint32_t  sv = sz + wb;              // Size of the outer left and right wall
  const float    &gx = p3db.GetGridSize().x; // X grid size
  const float    &gy = p3db.GetGridSize().y; // Y grid size
  const float    &gz = p3db.GetGridSize().z; // Z grid size
  const float    &tx = ox;                   // X translation factor
  const float     ty = oy + sy*gy;           // Y translation factor
  const float     tz = oz - sz*gz;           // Z translation factor
  const float     tl = tx - wl*gx;           // Left wall translation factor
  const float     tr = tx + sx*gx;           // Right wall translation factor
  const float     tb = oz + wb*gz;           // Back wall translation factor
  const float     tv = tr + wr*gx;           // Outer right wall translation factor
  const float     tk = ox + sx*gx + wr*gx;   // Outer back wall translation factor

  // Delete all previously created planes if asked
  if(deleteExistingPlanes) p3db.ClearAll();

  // Generate the floor
  p3db.GeneratePlane(np+"floor-main", _F(ox, oy, oz), sz, sx, EPD_0P0);
  if(ft > 0) {
    p3db.GeneratePlane(np+"floor-front", _F(tl, oy, tz), sy, sb, EPD_00N);
    if(fw) {
      p3db.GeneratePlane(np+"floor-left", _F(tl, oy, tb), sy, sv, EPD_N00);
      p3db.GeneratePlane(np+"floor-back-wall", _F(tk, oy, tb), sy, sb, EPD_00P);
      p3db.GeneratePlane(np+"floor-right", _F(tv, oy, tz), sy, sv, EPD_P00);
    }
  }

  // Generate the left wall
  if(lwt >= 0) {
    p3db.GeneratePlane(np+"left-in-wall", _F(ox, ty, tz), sy, sz, EPD_P00);
    if(wl) {
      p3db.GeneratePlane(np+"left-front-wall", _F(tl, ty, tz), sy, wl, EPD_00N);
      p3db.GeneratePlane(np+"left-top-wall", _F(tl, ty, oz), sz, wl, EPD_0P0);
    }
    if(fw) {
      p3db.GeneratePlane(np+"left-out-wall", _F(tl, ty, tb), sy, sv, EPD_N00);
      if(bwt < 0) p3db.GeneratePlane(np+"left-back-wall", _F(ox, ty, oz), sy, wl, EPD_00P);
    }
  }

  // Generate the back wall
  if(bwt >= 0) {
    p3db.GeneratePlane(np+"back-in-wall", _F(ox, ty, oz), sy, sx, EPD_00N);
    if(wb) p3db.GeneratePlane(np+"back-top-wall", _F(tl, ty, tb), wb, sb, EPD_0P0);
    if(lwt < 0) p3db.GeneratePlane(np+"back-left-wall", _F(ox, ty, tb), sy, wb, EPD_N00);
    if(rwt < 0) p3db.GeneratePlane(np+"back-right-wall", _F(tr, ty, oz), sy, wb, EPD_P00);
    if(fw) p3db.GeneratePlane(np+"back-out-wall", _F(tk, ty, tb), sy, sb, EPD_00P);
  }

  // Generate the right wall
  if(rwt >= 0) {
    p3db.GeneratePlane(np+"right-in-wall", _F(tr, ty, oz), sy, sz, EPD_N00);
    if(wr) {
      p3db.GeneratePlane(np+"right-front-wall", _F(tr, ty, tz), sy, wr, EPD_00N);
      p3db.GeneratePlane(np+"right-top-wall",  _F(tr, ty, oz), sz, wr, EPD_0P0);
    }
    if(fw) {
      p3db.GeneratePlane(np+"right-out", _F(tv, ty, tz), sy, sv, EPD_P00);
      if(bwt < 0) p3db.GeneratePlane(np+"right-back-wall", _F(tv, ty, oz), sy, wr, EPD_00P);
    }
  }
}

} // namespace vge
