/** \file Font.h */
// ############################################################################################
//  Font.h
//  Prototype & implementation of CFontDesc class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___FONT_H_
#define _VIDEO_ENVIRONMENT___FONT_H_

// VortexGE header
#include "../Base/SpecialClasses.h"


namespace vge {

  //!
  //! \brief The font description class
  //!
  //! A font description class used for easier font usage in video environment.
  //!
  //! \api beta
  class VGE_EXPORT CFontDesc {
      public:
        // Implement enum class wrappers for font styles
        VCLASS_ENUM_WRAPPER_IMPLEMENT(Pitch,     Any,  Monospace, Proportional)
        VCLASS_ENUM_WRAPPER_IMPLEMENT(Weight,    Any,  Regular,   Bold        )
        VCLASS_ENUM_WRAPPER_IMPLEMENT(Slant,     Any,  Normal,    Italic      )
        VCLASS_ENUM_WRAPPER_IMPLEMENT(Underline, None, Single,    Double      )

        // Default fonts
        static const CFontDesc DefaultSmallFont;
        static const CFontDesc DefaultBigFont;
        static const CFontDesc DefaultUtf8SmallFont;
        static const CFontDesc DefaultUtf8BigFont;

      public:
        //!
        //! Construct a new font class.
        //! \param[in] foundry   The font foundry, examples: "adobe", "bitstream", "ibm", etc. Fill with "*" for maximum flexibility.
        //! \param[in] family    The font family name (the face name), examples: "courier", "mincho", "times", etc.
        //! \param[in] pointSize The font point size(s), examples: "8", "12", "24", "8/12", "12/10/11", etc.
        //! \param[in] pitch     The font pitch, most fonts are Font::Pitch::Proportional.
        //! \param[in] weight    The font weight.
        //! \param[in] slant     The font slant.
        //! \param[in] underline The font underline method.
        //! \param[in] registry  The font registry, examples: "iso8859", "iso10646", "jisx0213.2000", etc. Fill with "*" for maximum flexibility (or "iso8859" if only need Latin character sets).
        inline CFontDesc(const string &foundry,
                         const string &family,
                         const string &pointSize,
                                Pitch pitch     =     Pitch::Any,
                               Weight weight    =    Weight::Any,
                                Slant slant     =     Slant::Any,
                            Underline underline = Underline::None,
                         const string &registry = "iso8859")
        : _foundry  (foundry),   _family  (family),  _pointSize(pointSize),
          _pitch    (pitch),     _weight  (weight),  _slant    (slant),
          _underline(underline), _registry(registry)
        {}
        //!
        //! Construct a new font class with ignoring the foundry.
        //! \n\n
        //! \note See the other constructor for some explanations about the parameters.
        inline CFontDesc(const string &family,
                         const string &pointSize,
                                Pitch pitch     =     Pitch::Any,
                               Weight weight    =    Weight::Any,
                                Slant slant     =     Slant::Any,
                            Underline underline = Underline::None,
                         const string &registry = "iso8859")
        : _foundry  ("*"),       _family  (family),  _pointSize(pointSize),
          _pitch    (pitch),     _weight  (weight),  _slant    (slant),
          _underline(underline), _registry(registry)
        {}

        // Getters
        const string &GetFoundry()   const { return(_foundry);   }
        const string &GetFamily()    const { return(_family);    }
        const string &GetPointSize() const { return(_pointSize); }
               Pitch GetPitch()      const { return(_pitch);     }
              Weight GetWeight()     const { return(_weight);    }
               Slant GetSlant()      const { return(_slant);     }
           Underline GetUnderline()  const { return(_underline); }
        const string &GetRegistry()  const { return(_registry);  }

        // Setters
        void SetFoundry  (const string &foundry  ) { _foundry   = foundry;   }
        void SetFamily   (const string &family   ) { _family    = family;    }
        void SetPointSize(const string &pointSize) { _pointSize = pointSize; }
        void SetPitch    (       Pitch pitch     ) { _pitch     = pitch;     }
        void SetWeight   (      Weight weight    ) { _weight    = weight;    }
        void SetSlant    (       Slant slant     ) { _slant     = slant;     }
        void SetUnderline(   Underline underline ) { _underline = underline; }
        void SetRegistry (const string &registry ) { _registry  = registry;  }

        // Helper functions
        //!
        //! Since a font description can contains multiple family names,
        //! use this function to parse them.
        void ParseFamilyNames(vector<string> &fName) const;
        //!
        //! Since a font description can contains multiple point sizes,
        //! use this function to parse them.
        void ParsePointSizes(vector<string> &pSize) const;

      protected:
          string _foundry;
          string _family;
          string _pointSize;
           Pitch _pitch;
          Weight _weight;
           Slant _slant;
       Underline _underline;
          string _registry;
  };

} // namespace vge
#endif


/**
 * \class vge::CFontDesc::Pitch
 * \brief An enumeration wrapper class for defining the font pitch
 *
 * Available values:
 * - CFontDesc::Pitch::Any
 * - CFontDesc::Pitch::Monospace
 * - CFontDesc::Pitch::Proportional
 *
 * \api beta
*/

/**
 * \class vge::CFontDesc::Weight
 * \brief An enumeration wrapper class for defining the font weight
 *
 * Available values:
 * - CFontDesc::Weight::Any
 * - CFontDesc::Weight::Regular
 * - CFontDesc::Weight::Bold
 *
 * \api beta
*/

/**
 * \class vge::CFontDesc::Slant
 * \brief An enumeration wrapper class for defining the font slant
 *
 * Available values:
 * - CFontDesc::Slant::Any
 * - CFontDesc::Slant::Normal
 * - CFontDesc::Slant::Italic
 *
 * \api beta
*/

/**
 * \class vge::CFontDesc::Underline
 * \brief An enumeration wrapper class for defining the font underline mode
 *
 * Available values:
 * - CFontDesc::Underline::None
 * - CFontDesc::Underline::Single
 * - CFontDesc::Underline::Double
 *
 * \api beta
*/
