// ############################################################################################
//  VX11Dialog4.cc
//  Implementation of dialog classes using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/DialogEnvironment.cc"


namespace vge {

// Delay constants used by text boxes in editing mode
const uint32_t TIME_DELAY   = 4000;                // Dispatch event delay ->   4 mS
const uint32_t CURSOR_DELAY = 100000 / TIME_DELAY; // Cursor blink delay   -> 100 mS


//
// Text boxes' sentry class
//
class _CTBSentry {
  public:
    inline _CTBSentry(bool &f, Display *d, GC gc, int32_t x1, int32_t y1, int32_t x2, int32_t y2, uint32_t ww, uint32_t wh)
    : _f(f), _d(d), _gc(gc)
    {
      _f = true;
      _cr[0].x = 0;    _cr[0].y = 0;    _cr[0].width = ww;    _cr[0].height = wh;
      _cr[1].x = 0;    _cr[1].y = 0;    _cr[1].width = x1;    _cr[1].height = wh;
      _cr[2].x = x1;   _cr[2].y = 0;    _cr[2].width = x2-x1; _cr[2].height = y1;
      _cr[3].x = x1;   _cr[3].y = y2+1; _cr[3].width = x2-x1; _cr[3].height = wh - y2;
      _cr[4].x = x2+1; _cr[4].y = 0;    _cr[4].width = ww-x2; _cr[4].height = wh;
    }
  
    inline ~_CTBSentry()
    {
      _f = false;
      DisableClipRegion();
    }
  
    inline void EnableClipRegion()
    { XSetClipRectangles(_d, _gc, 0, 0, &_cr[1], 4, Unsorted); }
  
    inline void DisableClipRegion()
    { XSetClipRectangles(_d, _gc, 0, 0, &_cr[0], 1, Unsorted); }

  private:
    bool      &_f;
    Display   *_d;
    GC         _gc;
    XRectangle _cr[5];
};


//
// Helper functions for drawing text boxes' text lines
//
#ifndef __USE_UTF8_DIALOG_STRING__

void _AsciiAdjustColPos_(size_t &curCol, TFont *font, const string &oldLine, const string &newLine)
{
  int         dr, as, ds;
  XCharStruct cs;
  XTextExtents(*font, oldLine.data(), curCol, &dr, &as, &ds, &cs);
  const size_t crw = cs.width;
  for(size_t i = 1; i <= newLine.length(); ++i) {
    XTextExtents(*font, newLine.data(), i, &dr, &as, &ds, &cs);
    if(size_t(cs.width) >= crw) {
      curCol = i;
      break;
    }
  }
}

#else

static int _Utf32CharExtents_(TFontSet *fontSet, const CUString::Utf32 &chr, XRectangle &overallLogicalReturn)
{
  string   utf8;
  CUString tmp; tmp += chr;
  tmp.ToUtf8(utf8);
  return(Xutf8TextExtents(*fontSet, &utf8[0], utf8.length(), NULL, &overallLogicalReturn));
}

static int _Utf32TextExtents_(TFontSet *fontSet, const CUString &str, XRectangle &overallLogicalReturn, size_t offset = 0, size_t length = 0)
{
  string utf8; 
  if(length)
    CUString::Utf32ToUtf8(utf8, &str[offset], length);
  else if(offset)
    CUString::Utf32ToUtf8(utf8, &str[offset], str.length() - offset);
  else
    str.ToUtf8(utf8);
  return(Xutf8TextExtents(*fontSet, &utf8[0], utf8.length(), NULL, &overallLogicalReturn));
}

static void _Utf32DrawString_(Display *dpy, Drawable drw, TFontSet *fontSet, GC &gc, int x, int y, const CUString &str, size_t offset = 0)
{
  string utf8;
  if(offset)
    CUString::Utf32ToUtf8(utf8, &str[offset], str.length() - offset);
  else
    str.ToUtf8(utf8);
  Xutf8DrawString(dpy, drw, *fontSet, gc, x, y, &utf8[0], utf8.length());
}

void _Utf32AdjustColPos_(size_t &curCol, TFontSet *fontSet, const CUString &oldLine, const CUString &newLine)
{
  XRectangle ttl;
  _Utf32TextExtents_(fontSet, oldLine, ttl, 0, curCol);
  const size_t crw = ttl.width;
  for(size_t i = 1; i <= newLine.length(); ++i) {
    _Utf32TextExtents_(fontSet, newLine, ttl, 0, i);
    if(ttl.width >= crw) {
      curCol = i;
      break;
    }
  }
}

#endif

#ifndef __USE_UTF8_DIALOG_STRING__
void _MLTBParseLines_(vector<string> &dst, const string &src)
#else
void _MLTBParseLines_(vector<CUString> &dst, const string &src)
#endif
{
  if(src.empty()) {
    dst.push_back(string(""));
    return;
  }

  string aLine;
  for(string::size_type i = 0; i < src.length(); ++i) {
    const char c = src[i];
    if(c == '\n') {
#ifndef __USE_UTF8_DIALOG_STRING__
      dst.push_back(aLine);
#else
      dst.push_back(CUString(aLine));
#endif
      aLine.clear();
    }
    else if(c != '\r') {
      aLine += c;
    }
  }
#ifndef __USE_UTF8_DIALOG_STRING__
  if(!aLine.empty()) dst.push_back(aLine);
#else
  if(!aLine.empty()) dst.push_back(CUString(aLine));
#endif 
}

#ifndef __USE_UTF8_DIALOG_STRING__
void _MLTBComposeLines_(string &dst, const vector<string> &src)
#else
void _MLTBComposeLines_(string &dst, const vector<CUString> &src)
#endif
{
#ifndef __USE_UTF8_DIALOG_STRING__
  typedef string StringT;
#else
  typedef CUString StringT;
#endif

  string compose;
  for(vector<StringT>::size_type i = 0; i < src.size(); ++i) {
    const StringT &aLine = src[i];
    if(!aLine.empty()) {
#ifndef __USE_UTF8_DIALOG_STRING__
      compose += aLine;
#else
      compose += aLine.ToUtf8();
#endif
      if(i != (src.size() - 1)) compose += '\n';
    }
    else
      compose += '\n';
  }
  swap(compose, dst);
}


//
// Simple-single-line-text-box control
//
CSLTextBox::~CSLTextBox() throw()
{
}

void CSLTextBox::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                             int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                             ETBType eType, bool bEscCancel, PFNSLTextBoxCallback pfnTBUserCallback)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_SLTextBox, n32X, n32Y, un32Width, un32Height);

  // Get the calculated coordinates
  int32_t n32X1, n32Y1, n32X2, n32Y2;
  n32X1 = _tBD.n32X1; n32X2 = _tBD.n32X2;
  n32Y1 = _tBD.n32Y1; n32Y2 = _tBD.n32Y2;

  // The face
  _tFace.x = n32X1 + 2;  _tFace.width  = (n32X2 - n32X1) - 3;
  _tFace.y = n32Y1 + 2;  _tFace.height = (n32Y2 - n32Y1) - 3;

  // The outline
  _tOutline.x = n32X1; _tOutline.width  = n32X2 - n32X1;
  _tOutline.y = n32Y1; _tOutline.height = n32Y2 - n32Y1;
  n32X1++; n32Y1++; n32X2--; n32Y2--;

  // The borders
  _atSegTL[0].x1 = n32X1;     _atSegTL[0].y1 = n32Y1;     _atSegTL[0].x2 = n32X2 - 1; _atSegTL[0].y2 = n32Y1;
  _atSegTL[1].x1 = n32X1;     _atSegTL[1].y1 = n32Y1 + 1; _atSegTL[1].x2 = n32X1;     _atSegTL[1].y2 = n32Y2;
  _atSegBR[0].x1 = n32X2;     _atSegBR[0].y1 = n32Y1;     _atSegBR[0].x2 = n32X2;     _atSegBR[0].y2 = n32Y2;
  _atSegBR[1].x1 = n32X1 + 1; _atSegBR[1].y1 = n32Y2;     _atSegBR[1].x2 = n32X2 - 1; _atSegBR[1].y2 = n32Y2;

  // The starting position for drawing text
  _un32Xt = n32X1 + 2;
#ifndef __USE_UTF8_DIALOG_STRING__
  _un32Yt = (n32Y1 + n32Y2 + (*vidEnv._tX11Style.ptFont)->ascent) / 2;
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, _DTM, strlen(_DTM), NULL, &tl);
  _un32Yt = (n32Y1 + n32Y2 - tl.y) / 2;
#endif

  // Set variables
  _eType      = eType;
  _bEscCancel = bEscCancel;
  _pfnTBCB    = pfnTBUserCallback;
  _inDoEdit   = false;
  _readOnly   = false;
}

void CSLTextBox::SetText(const char *pucNewText)
{
#ifndef __USE_UTF8_DIALOG_STRING__
  _csBuffer.clear();
  for(size_t i = 0; i < strlen(pucNewText); ++i) {
    char c = pucNewText[i];
    if(_validateChar(_csBuffer, c))
      _csBuffer += c;
    else
      VError(VERR_InvalidParameter, _NFL_, "Invalid character(s) found in the text");
  }
#else
  CUString src(pucNewText, CUString::ESCE_Utf8), dst;
  for(size_t i = 0; i < src.length(); ++i) {
    CUString::Utf32 c = src[i];
    if(_validateChar(_csBuffer, c))
      dst += c;
    else
      VError(VERR_InvalidParameter, _NFL_, "Invalid character(s) found in the text");
  }
  dst.ToUtf8(_csBuffer);
#endif
}

#ifndef __USE_UTF8_DIALOG_STRING__
bool CSLTextBox::_confirmAction(ETBEvent eventCode,
                                string &editBuffer,
                                string *pasteBuffer, char &charCode, 
                                EKeySymbol keySym)
#else
bool CSLTextBox::_confirmAction(ETBEvent eventCode,
                                CUString &editBuffer,
                                CUString *pasteBuffer, CUString::Utf32 &charCode,
                                EKeySymbol keySym)
#endif
{
  // Just return 'true' if there is no user callback function set or
  // the text is read-only
  if(!_pfnTBCB || _readOnly) return(true);

  bool ret;
  try {
    ret = _pfnTBCB(eventCode, *_tBD.pcsVidEnv, *this, editBuffer, pasteBuffer, charCode, keySym);
  }
  catch(...) {
    // Better assume that the confirmation is accepted rather than looping
    // forever because the user callback function throw exception
    ret = true;
  }

  return(ret);
}

#ifndef __USE_UTF8_DIALOG_STRING__
bool CSLTextBox::_validateChar(const string &editBuffer, char &charCode)
#else
bool CSLTextBox::_validateChar(const CUString &editBuffer, CUString::Utf32 &charCode)
#endif
{
  switch(_eType) {
    case ETBT_TextNormal:
      break;
    case ETBT_TextLowerCase:
      charCode = tolower(charCode);
      break;
    case ETBT_TextUpperCase:
      charCode = toupper(charCode);
      break;
    case ETBT_AZ09SNormal:
      if(!isalnum(charCode) && charCode!=' ') return(false);
      break;
    case ETBT_AZ09SLowerCase:
      if(!isalnum(charCode) && charCode!=' ') return(false);
      charCode = tolower(charCode);
      break;
    case ETBT_AZ09SUpperCase:
      if(!isalnum(charCode) && charCode!=' ') return(false);
      charCode = toupper(charCode);
      break;
    case ETBT_AZ09Normal:
      if(!isalnum(charCode)) return(false);
      break;
    case ETBT_AZ09LowerCase:
      if(!isalnum(charCode)) return(false);
      charCode = tolower(charCode);
      break;
    case ETBT_AZ09UpperCase:
      if(!isalnum(charCode)) return(false);
      charCode = toupper(charCode);
      break;
    case ETBT_Hexadecimal:
      if(!isxdigit(charCode)) return(false);
      charCode = toupper(charCode);
      break;
    case ETBT_Integer:
      if(charCode=='-' || charCode=='+')
        return(!editBuffer.length());
      else 
        return(isdigit(charCode));
      break;
    case ETBT_PositiveInteger:
      if(charCode == '+')
        return(!editBuffer.length());
      else 
        return(isdigit(charCode));
      break;
    case ETBT_Float:
      if(charCode=='-' || charCode=='+') {
        if(editBuffer.length())
          return(editBuffer[editBuffer.length() - 1] == 'e');
      }
      else if(toupper(charCode) == 'E') {
        charCode = 'e';
        return(editBuffer.find('e') == string::npos);
      }
      else if(charCode=='.' || charCode==',') {
        charCode = CBasicAppObjects::DecimalSeparator();
        return(editBuffer.find(CBasicAppObjects::DecimalSeparator())==string::npos &&
               editBuffer.find('e')==string::npos);
      }
      else 
        return(isdigit(charCode));
      break;
    case ETBT_PositiveFloat:
      if(charCode=='-' || charCode=='+') {
        if(editBuffer.length())
          return(editBuffer[editBuffer.length() - 1] == 'e');
        else
          return(charCode != '-');
      }
      else if(toupper(charCode) == 'E') {
        charCode = 'e';
        return(editBuffer.find('e') == string::npos);
      }
      else if((charCode=='.' || charCode==',')) {
        charCode = CBasicAppObjects::DecimalSeparator();
        return(editBuffer.find(CBasicAppObjects::DecimalSeparator())==string::npos &&
               editBuffer.find('e')==string::npos);
      }
      else 
        return(isdigit(charCode));
      break;
    default: break;
  };
  return(true);
}

bool CSLTextBox::DoEdit(CEventHandlerBase *propagateTo)
{
  const uint32_t KEYREP_START = TKeySettings::GLOBAL_SET.keyRepeatStart * 1000 / TIME_DELAY;
  const uint32_t KEYREP_DELAY = TKeySettings::GLOBAL_SET.keyRepeatDelay * 1000 / TIME_DELAY;

  // Select the video environment to be used
  CVideoEnv *svetbu = propagateTo ? propagateTo->HookedVideoEnv() : _tBD.pcsVidEnv;
  CVideoEnv &vidEnv = *svetbu;

  // For convenience
#ifndef __USE_UTF8_DIALOG_STRING__
  typedef char   CharT;
  typedef string StringT;
#else
  typedef CUString::Utf32 CharT;
  typedef CUString        StringT;
#endif

  // If disabled then the text can't be edited
  if(!_tBD.bEnabled) return(false);

  // Instantiate the sentry class
  _CTBSentry _sentry(_inDoEdit, vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC,
                     _tBD.n32X1, _tBD.n32Y1, _tBD.n32X2, _tBD.n32Y2,
                     vidEnv._tX11.un32Width, vidEnv._tX11.un32Height);

  // Copy the text to edit buffer
  StringT buff(_csBuffer);

  // Store the current view
  CVideoStorage vStorage;
  vStorage.Initialize(vidEnv, 1);
  vStorage.StoreCurrentView();

  // For event propagation
  void *eventStorage = NULL;

  // Used for editing
  int32_t dPos = 0;
  int32_t aWidth;
#ifndef __USE_UTF8_DIALOG_STRING__
  aWidth = _tFace.width - XTextWidth(*vidEnv._tX11CS.ptFont, "W", 1) + 1;
#else 
  {
    XRectangle wtl;
    Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, "W", 1, NULL, &wtl);
    aWidth = _tFace.width - wtl.width + 1;
  }
#endif

  // Editing loop
  bool retVal = false; // Flag to indicate if the editting is accepted
  try {
    int32_t    posX          = 0;           // Starting coordinate for drawing text
    uint32_t   curCnt        = 0;           // Counter for cursor blink
    bool       eLEM          = false;       // Flag for leaving the editing mode (as if we pressed the enter key)
    bool       bPress        = false;       // Flag to restore display after mouse button press
    bool       curShow       = true;        // Flag for cursor blink
    bool       redrawText    = true;        // Flag for redrawing text
    bool       updateDisplay = true;        // Flag for updating display
    bool       ctrl          = false;       // Flag for control key
    bool       kyRep         = false;       // Flag for repetitive key
    uint32_t   kyCnt         = 0;           // Counter for repetitive key
    EKeySymbol eKSym         = VKS_UNKNOWN; // Pressed key's symbol
    CharT      chAdd         = 0;           // Character to be added
    while(1) {
      // Update display if needed
      if(updateDisplay && vidEnv.IsActive()) {
        updateDisplay = false;
        // Redraw text
        if(redrawText) {
          redrawText = false;
          // Calculate starting index for displaying text
#ifndef __USE_UTF8_DIALOG_STRING__
          if(XTextWidth(*vidEnv._tX11CS.ptFont, &buff[0], buff.length()) >= aWidth) {
            for(size_t i = 0; i < buff.length(); ++i) {
              if(XTextWidth(*vidEnv._tX11CS.ptFont, &buff[i], buff.length() - i) < aWidth) {
#else
          XRectangle tl;
          _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, buff, tl);
          if(tl.width >= aWidth) {
            for(size_t i = 0; i < buff.length(); ++i) {
              _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, buff, tl, i);
              if(tl.width < aWidth) {
#endif
                dPos = i;
                break;
              }
            }
            if(dPos >= 2) dPos -= 2;
#ifndef __USE_UTF8_DIALOG_STRING__
            posX = int32_t(_un32Xt + aWidth) - XTextWidth(*vidEnv._tX11CS.ptFont, &buff[dPos], buff.length() - dPos);
#else
            _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, buff, tl, dPos);
            posX = int32_t(_un32Xt + aWidth) - tl.width;
#endif
          }
          else {
            dPos = 0;
            posX = _un32Xt;
          }
          // Empty the text area
          vidEnv.SetFillStyle(FILL_SOLID);
          if(_readOnly) vidEnv.SetForegroundColor(vidEnv._tX11Style.xFaceLocked);
          else          vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
          XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                         _tFace.x, _tFace.y, _tFace.width, _tFace.height);
          // Set the clip region
          XRectangle rClip = _tFace;
          rClip.width  -= 2;
          rClip.height -= 2;
          XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &rClip, 1, Unsorted);
          // Draw the text
          vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
          vidEnv.SetFont(vidEnv._tX11Style.ptFont);
          XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                      posX, _un32Yt, &buff[dPos], buff.length() - dPos);
#else
          _Utf32DrawString_(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                            vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                            posX, _un32Yt, buff, dPos);
#endif
          // Restore the clip region
          XRectangle clip;
          clip.x = 0; clip.y = 0; clip.width = vidEnv._tX11.un32Width; clip.height = vidEnv._tX11.un32Height;
          XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &clip, 1, Unsorted);
        }
        // Draw cursor
#ifndef __USE_UTF8_DIALOG_STRING__
        uint32_t lX = posX + XTextWidth(*vidEnv._tX11CS.ptFont, &buff[dPos], buff.length() - dPos);
#else
        XRectangle tl;
        _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, buff, tl, dPos);
        uint32_t lX = posX + tl.width;
#endif
        if(curShow)
          vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
        else
          vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
        XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                       lX, _tFace.y + 2, 2, _tFace.height - 4);
        // Update display
        vidEnv.UpdateDisplay();
      }
      // Update the cursor appearance
      curCnt++;
      if(curCnt >= CURSOR_DELAY) {
        updateDisplay = true;
        curShow       = !curShow;
        curCnt        = 0;
      }
      // Check event
      usleep(TIME_DELAY);
      vidEnv.DispatchEvent();
      if(vidEnv.IsInactive()) {
        usleep(100000);
        continue;
      }
      else if(vidEnv.CheckEvent()) {
        // Expose event
        if(vidEnv.EventExpose())
          updateDisplay = true;
        else if(!_readOnly && vidEnv.EventPasteData()) {
#ifndef __USE_UTF8_DIALOG_STRING__
          string sget;
          CClipboard::CopyFrom(sget);
#else
          string stmp;
          CClipboard::CopyFrom(stmp);
          CUString sget(stmp, CUString::ESCE_Utf8);
#endif
          if(!sget.empty()) {
            CharT zero = 0;
            if(_confirmAction(ETBE_Paste, buff, &sget, zero, VKS_UNKNOWN)) {
              buff += sget;
              updateDisplay = true;
              redrawText    = true;
            }
          }
        }
        // Keyboard events
        else if(vidEnv.EventKeyPress()) {
          EKeySymbol eKey = vidEnv.GetKeySym();
          if((eKey == VKS_ENTER) || (eKey == VKS_KP_ENTER)) {
            eLEM  = true;
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
          }
          else if((eKey == VKS_LCTRL) || (eKey == VKS_RCTRL)) {
            ctrl  = true;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
          }
          else if(eKey == VKS_ESC) {
            // Set flags
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
            // Set flags
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
            // Propagate the event
            if(propagateTo) {
              vidEnv.PutBackEvent();
              vidEnv.PopEvent(&eventStorage);
            }
            // Execute the confirmation function
            const ETBEvent evcd = _bEscCancel ? ETBE_ConfirmCancel : ETBE_ConfirmOK;
                  CharT    zero = 0;
            if(_confirmAction(evcd, buff, NULL, zero, VKS_ESC)) {
              retVal = true;
              break;
            }
            else {
              vidEnv.FreeStoredEvent(&eventStorage);
              updateDisplay = true;
            }
          }
          else if(!_readOnly && (eKey == VKS_BS)) {
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_BS;
          }
          else if(eKey == VKS_TAB) {
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
          }
          else if(!_readOnly && ctrl && (eKey == 'v')) {
            CClipboard::RequestData(CClipboard::CDT_String);
          }
          else if(ctrl && (eKey == 'c')) {
#ifndef __USE_UTF8_DIALOG_STRING__
            CClipboard::CopyTo(buff);
#else
            CClipboard::CopyTo(buff.ToUtf8());
#endif
          }
          else if(!_readOnly) {
#ifndef __USE_UTF8_DIALOG_STRING__
            char kb[32];
            if(vidEnv.GetKeyString(kb, sizeof(kb))) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = kb[0]; eKSym = eKey;
            }
#else
            char     kb[32];
            uint16_t ucs;
            uint32_t rz = vidEnv.GetKeyString(kb, sizeof(kb), &ucs);
            if(ucs) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = ucs;   eKSym = eKey;
            }
            else if(rz) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = kb[0]; eKSym = eKey;
            }
#endif
          }
        }
        else if(vidEnv.EventKeyRelease()) {
          EKeySymbol eKey = vidEnv.GetKeySym();

          if((eKey == VKS_LCTRL) || (eKey == VKS_RCTRL))
            ctrl = false;

          kyRep = false; kyCnt = 0;
          chAdd = 0;     eKSym = VKS_UNKNOWN;
        }
        // Mouse events
        else if(vidEnv.EventButtonPress()) {
          // Propagate the mouse event
          if(propagateTo && !MouseOver(vidEnv.GetButtonX(), vidEnv.GetButtonY())) {
            vidEnv.PutBackEvent();
            _sentry.EnableClipRegion();
            propagateTo->HandleEvents();
            _sentry.DisableClipRegion();
            bPress = true;
          }
        }
        else if(vidEnv.EventMouseMove()) {
          // Propagate the mouse event
          if(propagateTo && !MouseOver(vidEnv.GetMouseX(), vidEnv.GetMouseY())) {
            vidEnv.PutBackEvent();
            _sentry.EnableClipRegion();
            propagateTo->HandleEvents();
            _sentry.DisableClipRegion();
          }
        }
        else if(vidEnv.EventButtonRelease() && !MouseOver(vidEnv.GetButtonX(), vidEnv.GetButtonY())) {
          eLEM = true;
        }
        // Close window event
        else if(vidEnv.EventCloseWindow()) {
          eLEM = true;
        }
        // Check if we should leave the editing mode
        // (as if we pressed the enter key)
        if(eLEM) {
          // Set flag
          eLEM = false;
          // Propagate the event
          if(propagateTo) {
            vidEnv.PutBackEvent();
            vidEnv.PopEvent(&eventStorage);
          }
          // Execute the confirmation function
          CharT zero = 0;
          bool  ret  = _confirmAction(ETBE_ConfirmOK, buff, NULL, zero, VKS_ENTER);
          // Exit if the confirmation is accepted
          if(ret) {
            retVal = true;
            break;
          }
          else {
            vidEnv.FreeStoredEvent(&eventStorage);
            if(bPress) vStorage.RestoreCurrentView();
            bPress        = false;
            updateDisplay = true;
          }
          // Set flags
          ctrl  = false;
          kyRep = false; kyCnt = 0;
          chAdd = 0;     eKSym = VKS_UNKNOWN;
        }
      }
      // Edit the text
      if(eKSym == VKS_BS) {
        if(!kyCnt && buff.length()) {
          if(_confirmAction(ETBE_DelChar, buff, NULL, buff[buff.length() - 1], VKS_BS)) {
            buff.erase(buff.length() - 1);
            updateDisplay = true;
            redrawText    = true;
          }
        }
        kyCnt++;
      }
      else if(chAdd) {
        if(!kyCnt) {
          if(_validateChar(buff, chAdd) && _confirmAction(ETBE_AddChar, buff, NULL, chAdd, eKSym)) {
            buff += chAdd;
            updateDisplay = true;
            redrawText    = true;
          }
        }
        kyCnt++;
      }
      // Check the key repeat delay counter
      if(kyRep) {
        if(kyCnt >= KEYREP_DELAY) kyCnt = 0;
      } 
      else if(kyCnt >= KEYREP_START) {
        kyRep = true;
        kyCnt = 0;
      }
    }
  }
  catch(...) {
    vStorage.RestoreCurrentView();
    throw;
  }

  // Restore the view
  vStorage.RestoreCurrentView();

  // Copy the editing result
#ifndef __USE_UTF8_DIALOG_STRING__
  if(retVal) _csBuffer = buff;
#else
  if(retVal) buff.ToUtf8(_csBuffer);
#endif

  // Propagate the pending event (if any)
  if(eventStorage) {
    vidEnv.PushEvent(&eventStorage);
    //propagateTo->HandleEvents();
  }

  // Return the result
  return(retVal);
}

void CSLTextBox::DrawControl() const
{
  DrawControl(false);
}

void CSLTextBox::DrawControl(bool bPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Don't redraw if in editing mode
  if(_inDoEdit) return;

  // Draw face
  vidEnv.SetFillStyle(FILL_SOLID);
  if(_readOnly) vidEnv.SetForegroundColor(vidEnv._tX11Style.xFaceLocked);
  else          vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tFace.x, _tFace.y, _tFace.width, _tFace.height);

  // Draw outline
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tOutline.x, _tOutline.y, _tOutline.width, _tOutline.height);

  // Draw border
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegTL), 2);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegBR), 2);

  // Set clip region
  XRectangle rClip = _tFace;
  rClip.width  -= 2;
  rClip.height -= 2;
  XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &rClip, 1, Unsorted);

  // Draw text
#ifndef __USE_UTF8_DIALOG_STRING__
  vidEnv.SetFont(vidEnv._tX11Style.ptFont);
#endif
  if(_tBD.bEnabled) {
    if(bPressed)
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
    else
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _un32Xt, _un32Yt, _csBuffer.data(), _csBuffer.length());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _un32Xt, _un32Yt, _csBuffer.data(), _csBuffer.length());
#endif
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _un32Xt, _un32Yt, _csBuffer.data(), _csBuffer.length());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _un32Xt, _un32Yt, _csBuffer.data(), _csBuffer.length());
#endif
  }

  // Restore clip region
  XRectangle clip;
  clip.x = 0; clip.y = 0; clip.width = vidEnv._tX11.un32Width; clip.height = vidEnv._tX11.un32Height;
  XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &clip, 1, Unsorted);
}

void CSLTextBox::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  DrawControl(MouseOver(n32MouseX, n32MouseY) & bPressed);
}


//
// Multi-line-text-box control
//
CMLTextBox::~CMLTextBox() throw()
{
}

void CMLTextBox::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                             int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                             bool bEscCancel, PFNMLTextBoxCallback pfnTBUserCallback)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_MLTextBox, n32X, n32Y, un32Width, un32Height);

  // Get the calculated coordinates
  int32_t n32X1, n32Y1, n32X2, n32Y2;
  n32X1 = _tBD.n32X1; n32X2 = _tBD.n32X2;
  n32Y1 = _tBD.n32Y1; n32Y2 = _tBD.n32Y2;

  // The face
  _tFace.x = n32X1 + 2;  _tFace.width  = (n32X2 - n32X1) - 3;
  _tFace.y = n32Y1 + 2;  _tFace.height = (n32Y2 - n32Y1) - 3;

  // The outline
  _tOutline.x = n32X1; _tOutline.width  = n32X2 - n32X1;
  _tOutline.y = n32Y1; _tOutline.height = n32Y2 - n32Y1;
  n32X1++; n32Y1++; n32X2--; n32Y2--;

  // The borders
  _atSegTL[0].x1 = n32X1;     _atSegTL[0].y1 = n32Y1;     _atSegTL[0].x2 = n32X2 - 1; _atSegTL[0].y2 = n32Y1;
  _atSegTL[1].x1 = n32X1;     _atSegTL[1].y1 = n32Y1 + 1; _atSegTL[1].x2 = n32X1;     _atSegTL[1].y2 = n32Y2;
  _atSegBR[0].x1 = n32X2;     _atSegBR[0].y1 = n32Y1;     _atSegBR[0].x2 = n32X2;     _atSegBR[0].y2 = n32Y2;
  _atSegBR[1].x1 = n32X1 + 1; _atSegBR[1].y1 = n32Y2;     _atSegBR[1].x2 = n32X2 - 1; _atSegBR[1].y2 = n32Y2;

  // The starting position for drawing text
  _un32Xt = n32X1 + 2;
  _un32Yt = n32Y1 + 2;

  // Set variables
  _bEscCancel = bEscCancel;
  _pfnTBCB    = pfnTBUserCallback;
  _inDoEdit   = false;
  _readOnly   = false;
}

void CMLTextBox::SetText(const char *pucNewText)
{
  _csBuffer = pucNewText;
}

#ifndef __USE_UTF8_DIALOG_STRING__
bool CMLTextBox::_confirmAction(ETBEvent eventCode,
                                vector<string> &editBuffer,
                                size_t actRow, size_t actCol,
                                vector<string> *pasteBuffer, char &charCode,
                                EKeySymbol keySym)
#else
bool CMLTextBox::_confirmAction(ETBEvent eventCode,
                                vector<CUString> &editBuffer,
                                size_t actRow, size_t actCol,
                                vector<CUString> *pasteBuffer, CUString::Utf32 &charCode,
                                EKeySymbol keySym)
#endif
{
  // Just return 'true' if there is no user callback function set or
  // the text is read-only
  if(!_pfnTBCB || _readOnly) return(true);

  bool ret;
  try {
    ret = _pfnTBCB(eventCode, *_tBD.pcsVidEnv, *this, editBuffer, actRow, actCol, pasteBuffer, charCode, keySym);
  }
  catch(...) {
    // Better assume that the confirmation is accepted rather than looping
    // forever because the user callback function throw exception
    ret = true;
  }

  return(ret);
}

bool CMLTextBox::DoEdit(CEventHandlerBase *propagateTo)
{
  const uint32_t KEYREP_START = TKeySettings::GLOBAL_SET.keyRepeatStart * 1000 / TIME_DELAY;
  const uint32_t KEYREP_DELAY = TKeySettings::GLOBAL_SET.keyRepeatDelay * 1000 / TIME_DELAY;

  // Select the video environment to be used
  CVideoEnv *svetbu = propagateTo ? propagateTo->HookedVideoEnv() : _tBD.pcsVidEnv;
  CVideoEnv &vidEnv = *svetbu;

  // For convenience
#ifndef __USE_UTF8_DIALOG_STRING__
  typedef char   CharT;
  typedef string StringT;
#else
  typedef CUString::Utf32 CharT;
  typedef CUString        StringT;
#endif

  // If disabled then the text can't be edited
  if(!_tBD.bEnabled) return(false);

  // Instantiate the sentry class
  _CTBSentry _sentry(_inDoEdit, vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC,
                     _tBD.n32X1, _tBD.n32Y1, _tBD.n32X2, _tBD.n32Y2,
                     vidEnv._tX11.un32Width, vidEnv._tX11.un32Height);

  // Parse the text to vector of lines
  vector<StringT> lines;
  _MLTBParseLines_(lines, _csBuffer);

  // Store the current view
  CVideoStorage vStorage;
  vStorage.Initialize(vidEnv, 1);
  vStorage.StoreCurrentView();

  // For event propagation
  void *eventStorage = NULL;

  // Editing loop
  bool retVal = false;// Flag to indicate if the editting is accepted
  try {
    size_t     dspRow        = 0;           // Starting position of the characters to be displayed
    size_t     dspCol        = 0;           // ---
    size_t     curRow        = 0;           // Current cursor position
    size_t     curCol        = 0;           // ---
    uint32_t   curCnt        = 0;           // Counter for cursor blink
    size_t     pgMv          = 10;          // Page up and Page down strength
    bool       eLEM          = false;       // Flag for leaving the editing mode (as if we pressed the enter key)
    bool       bPress        = false;       // Flag to restore display after mouse button press
    bool       curShow       = true;        // Flag for cursor blink
    bool       updateDisplay = true;        // Flag for updating display
    bool       ctrl          = false;       // Flag for control key
    bool       overwrite     = false;       // Flag for overwriting characters
    bool       kyRep         = false;       // Flag for repetitive key
    uint32_t   kyCnt         = 0;           // Counter for repetitive key
    EKeySymbol eKSym         = VKS_UNKNOWN; // Pressed key's symbol
    CharT      chAdd         = 0;           // Character to be added
    while(1) {
      // Update display if needed
      if(updateDisplay && vidEnv.IsActive()) {
        updateDisplay = false;
        // Empty the text area
        vidEnv.SetFillStyle(FILL_SOLID);
        if(_readOnly) vidEnv.SetForegroundColor(vidEnv._tX11Style.xFaceLocked);
        else          vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
        XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                       _tFace.x, _tFace.y, _tFace.width, _tFace.height);
        // Prepare the GC for drawing text
        vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
        vidEnv.SetFont(vidEnv._tX11Style.ptFont);
#endif
        // Set the clip region for drawing text
        XRectangle rClip = _tFace;
        rClip.width  -= 2;
        rClip.height -= 2;
        XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &rClip, 1, Unsorted);
        // Calculate starting height using dummy text
#ifndef __USE_UTF8_DIALOG_STRING__
        int         dr, as, ds;
        XCharStruct cs;
        XTextExtents(*vidEnv._tX11Style.ptFont, _DTM, strlen(_DTM), &dr, &as, &ds, &cs);
#else
        XRectangle tl;
        size_t     defH;
        Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, _DTM, strlen(_DTM), NULL, &tl);
        defH = tl.height;
#endif
        // Adjust the display's row & column positions
        if(curRow <= dspRow) dspRow = curRow ? (curRow - 1) : 0;
        if(curCol <= dspCol) dspCol = curCol ? (curCol - 1) : 0;
        // Calculate the X coordinate substraction factor for drawing text
#ifndef __USE_UTF8_DIALOG_STRING__
        XTextExtents(*vidEnv._tX11Style.ptFont, lines[curRow].data(), dspCol, &dr, &as, &ds, &cs);
        const ssize_t xneg = cs.width;
#else
        if(dspCol)
          _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, lines[curRow], tl, 0, dspCol);
        else
          tl.width = 0;
        const ssize_t xneg = tl.width;
#endif
        // Draw the text and cursor
        ssize_t cx = -1;      // Cursor's X coordinate
        ssize_t cy = -1;      // Cursor's Y coordinate
        ssize_t ch = -1;      // Cursor's height
        ssize_t y  = _un32Yt; // Y coordinate for drawing text
        for(size_t i = dspRow; i < lines.size(); ++i) {
          // For convenience
          const StringT &aLine = lines[i];
          // Save the cursor Y coordinate
          if(i == curRow) cy = y;
          // Draw the text
          if(aLine.length() <= dspCol) {
#ifndef __USE_UTF8_DIALOG_STRING__
            y += (as + ds);
#else
            y += tl.height;
#endif
          }
          else {
#ifndef __USE_UTF8_DIALOG_STRING__
            XTextExtents(*vidEnv._tX11Style.ptFont, aLine.data(), aLine.length(), &dr, &as, &ds, &cs);
            //XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
            //            _un32Xt, y + as, aLine.data() + dspCol, aLine.length() - dspCol);
            XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                        _un32Xt - xneg, y + as, aLine.data(), aLine.length());
            y += (as + ds);
#else
            if(aLine.empty())
              tl.height = defH;
            else {
              _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, aLine, tl);
              //_Utf32DrawString_(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
              //                  vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
              //                  _un32Xt, y - tl.y, aLine, dspCol);
              _Utf32DrawString_(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                                vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                                _un32Xt - xneg, y - tl.y, aLine);
            }
            y += tl.height;
#endif
          }
          // Calculate the cursor X coordinate and the cursor height
          if(i == curRow) {
            curCol = min<size_t>(aLine.length(), curCol);
#ifndef __USE_UTF8_DIALOG_STRING__
            int         cdr, cas, cds;
            XCharStruct ccs;
            //XTextExtents(*vidEnv._tX11Style.ptFont, aLine.data() + dspCol, curCol - dspCol, &cdr, &cas, &cds, &ccs);
            XTextExtents(*vidEnv._tX11Style.ptFont, aLine.data(), curCol, &cdr, &cas, &cds, &ccs);
            //cx = _un32Xt + ccs.width;
            cx = ssize_t(_un32Xt + ccs.width) - xneg;
            ch = as + ds;
#else
            if(aLine.empty() || !curCol) {
              cx = _un32Xt;
              ch = defH;
            }
            else {
              XRectangle ctl;
              //_Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, aLine, ctl, dspCol, curCol - dspCol);
              _Utf32TextExtents_(vidEnv._tX11Style.ptFontSet, aLine, ctl, 0, curCol);
              //cx = _un32Xt + ctl.width;
              cx = ssize_t(_un32Xt + ctl.width) - xneg;
              ch = ctl.height;
            }
#endif
          }
        }
        // Draw cursor
        if(curShow && ch>=0) {
          vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
          if(overwrite) {
            size_t      pos = min<size_t>(curCol, lines[curRow].length() - 1);
#ifndef __USE_UTF8_DIALOG_STRING__
            int         tdr, tas, tds;
            XCharStruct tcs;
            XTextExtents(*vidEnv._tX11Style.ptFont, &lines[curRow][pos], 1, &tdr, &tas, &tds, &tcs);
            XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                           cx, cy, tcs.width, ch); //cx, cy+ch-1, tcs.width, 2);
#else
            XRectangle ttl;
            _Utf32CharExtents_(vidEnv._tX11Style.ptFontSet, lines[curRow][pos], ttl);
            XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                           cx, cy, ttl.width, ch);
#endif
          }
          else
            XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                           cx, cy, 2, ch);
        }
        // Adjust the display's row & column positions
#ifndef __USE_UTF8_DIALOG_STRING__
        const ssize_t yl = (rClip.y + rClip.height - as - ds);
#else
        const ssize_t yl = (rClip.y + rClip.height - tl.height);
#endif
        if(cy >= yl) {
          const size_t f = ((cy - yl) / ch) + 1;
          dspRow += f;
        }
        const ssize_t xf = ch / 3;
        const ssize_t xl = (rClip.x + rClip.width - xf);
        if(cx >= xl) {
          const size_t f = ((cx - xl) / xf) + 1;
          dspCol += f;
        }
        // Restore the clip region
        XRectangle clip;
        clip.x = 0; clip.y = 0; clip.width = vidEnv._tX11.un32Width; clip.height = vidEnv._tX11.un32Height;
        XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &clip, 1, Unsorted);
        // Update display
        vidEnv.UpdateDisplay();
      }
      // Update the cursor appearance
      curCnt++;
      if(curCnt >= CURSOR_DELAY) {
        updateDisplay = true;
        curShow       = !curShow;
        curCnt        = 0;
      }
      // Check event
      usleep(TIME_DELAY);
      vidEnv.DispatchEvent();
      if(vidEnv.IsInactive()) {
        usleep(100000);
        continue;
      }
      else if(vidEnv.CheckEvent()) {
        // Expose event
        if(vidEnv.EventExpose())
          updateDisplay = true;
        // Paste data event
        else if(!_readOnly && vidEnv.EventPasteData()) {
          string sget;
          CClipboard::CopyFrom(sget);
          if(!sget.empty()) {
            CharT           zero = 0;
            vector<StringT> tls;
            _MLTBParseLines_(tls, sget);
            if(_confirmAction(ETBE_Paste, lines, curRow, curCol, &tls, zero, VKS_UNKNOWN)) {
              // At the beginning of a line
              if(curCol == 0) {
                const u_char sf = (sget[sget.length() - 1] == '\n') ? 0 : 1;
                for(vector<StringT>::size_type i = 0; i < (tls.size() - sf); ++i) {
                  lines.insert(lines.begin() + curRow, tls[i]);
                  curRow++;
                }
                // Does the last line to be pasted not end with '\n' ?
                if(sf) {
                  lines[curRow] = tls[tls.size() - 1] + lines[curRow];
                  curCol = tls[tls.size() - 1].length();
                }
                updateDisplay = true;
              }
              // At the end of a line
              else if(curCol == lines[curRow].length()) {
                lines[curRow] += tls[0];
                curRow++;
                for(vector<StringT>::size_type i = 1; i < tls.size(); ++i) {
                  lines.insert(lines.begin() + curRow, tls[i]);
                  curRow++;
                }
                // Does the last line to be pasted end with '\n' ?
                if(sget[sget.length() - 1] == '\n') {
                  lines.insert(lines.begin() + curRow,  StringT());
                  curCol = 0;
                }
                // Does not end with '\n'
                else {
                  curRow--;
                  curCol = lines[curRow].length();
                }
                updateDisplay = true;
              }
              // In the middle of a line
              else {
                const u_char sf = (sget[sget.length() - 1] == '\n') ? 0 : 1;
                if(sf && (tls.size() == 1)) {
                  lines[curRow].insert(curCol, tls[0]);
                }
                else {
                  const StringT str1 = lines[curRow].substr(0, curCol) + tls[0];
                  const StringT str2 = sf
                                       ? tls[tls.size() - 1] + lines[curRow].substr(curCol, string::npos)
                                       : lines[curRow].substr(curCol, string::npos);
                  lines[curRow] = str1; 
                  curRow++;
                  if(tls.size() >= 3) {
                    for(vector<StringT>::size_type i = 1; i < (tls.size() - sf); ++i) {
                      lines.insert(lines.begin() + curRow, tls[i]);
                      curRow++;
                    }
                    // Does the last line to be pasted end with '\n' and not empty ?
                    if(!sf && !tls[tls.size() - 2].empty()) {
                      lines.insert(lines.begin() + curRow, tls[tls.size() - 2]);
                      curRow++;
                    }
                  }
                  lines.insert(lines.begin() + curRow, str2);
                  if(sf) curCol = tls[tls.size() - 1].length();
                  else   curCol = 0;
                  updateDisplay = true;
                }
              }
            }
          }
        }
        // Keyboard events
        else if(vidEnv.EventKeyPress()) {
          EKeySymbol eKey = vidEnv.GetKeySym();

          if((eKey == VKS_LEFT) || (eKey == VKS_KP_4)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_LEFT;
          }
          else if((eKey == VKS_RIGHT) || (eKey == VKS_KP_6)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_RIGHT;
          }
          else if((eKey == VKS_UP) || (eKey == VKS_KP_8)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UP;
          }
          else if((eKey == VKS_DOWN) || (eKey == VKS_KP_2)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_DOWN;
          }
          else if((eKey == VKS_PGUP) || (eKey == VKS_KP_9)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_PGUP;
          }
          else if((eKey == VKS_PGDN) || (eKey == VKS_KP_3)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_PGDN;
          }
          else if((eKey == VKS_HOME) || (eKey == VKS_KP_7)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
  
            curCnt = 999;
            if(ctrl) curRow = 0;
            curCol = 0;
          }
          else if((eKey == VKS_END) || (eKey == VKS_KP_1)) {
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
  
            curCnt = 999;
            if(ctrl) curRow = lines.size() - 1;
            curCol = lines[curRow].length();
          }
          else if((eKey == VKS_ENTER) || (eKey == VKS_KP_ENTER)) {
            // ctrl-enter will make us leave the editing mode
            eLEM = ctrl;
            // Process further if not read only
            if(!_readOnly) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = 0;     eKSym = VKS_ENTER;
            }
          }
          else if(!_readOnly && (eKey == VKS_BS)) {
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_BS;
            
          }
          else if(!_readOnly && ((eKey == VKS_DEL) || (eKey == VKS_KP_PERIOD))) {
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_DEL;
          }
          else if((eKey == VKS_LCTRL) || (eKey == VKS_RCTRL)) {
            ctrl  = true;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
          }
          else if(eKey == VKS_ESC) {
            // Set flags
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
            // Propagate the event
            if(propagateTo) {
              vidEnv.PutBackEvent();
              vidEnv.PopEvent(&eventStorage);
            }
            // Execute the confirmation function
            const ETBEvent evcd = _bEscCancel ? ETBE_ConfirmCancel : ETBE_ConfirmOK;
                  CharT    zero = 0;
            if(_confirmAction(evcd, lines, curRow, curCol, NULL, zero, VKS_ESC)) {
              retVal = true;
              break;
            }
            else {
              vidEnv.FreeStoredEvent(&eventStorage);
              updateDisplay = true;
            }
            // Set flags
            ctrl  = false;
            kyRep = false; kyCnt = 0;
            chAdd = 0;     eKSym = VKS_UNKNOWN;
          }
          else if(!_readOnly && ctrl && (eKey == 'v')) {
            CClipboard::RequestData(CClipboard::CDT_String);
          }
          else if(ctrl && (eKey == 'c')) {
            string tmp;
            _MLTBComposeLines_(tmp, lines);
            CClipboard::CopyTo(tmp);
          }
          else if(!_readOnly) {
#ifndef __USE_UTF8_DIALOG_STRING__
            char kb[32];
            if(vidEnv.GetKeyString(kb, sizeof(kb))) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = kb[0]; eKSym = eKey;
            }
#else
            char     kb[32];
            uint16_t ucs;
            uint32_t rz = vidEnv.GetKeyString(kb, sizeof(kb), &ucs);
            if(ucs) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = ucs;   eKSym = eKey;
            }
            else if(rz) {
              ctrl  = false;
              kyRep = false; kyCnt = 0;
              chAdd = kb[0]; eKSym = eKey;
            }
#endif
          }
        }
        else if(vidEnv.EventKeyRelease()) {
          EKeySymbol eKey = vidEnv.GetKeySym();

          if((eKey == VKS_INS) || (eKey == VKS_KP_0))
            overwrite = !overwrite;
          else if((eKey == VKS_LCTRL) || (eKey == VKS_RCTRL))
            ctrl = false;
  
          kyRep = false; kyCnt = 0;
          chAdd = 0;     eKSym = VKS_UNKNOWN;
        }
        // Mouse events
        else if(vidEnv.EventButtonPress()) {
          // Propagate the mouse event
          if(propagateTo && !MouseOver(vidEnv.GetButtonX(), vidEnv.GetButtonY())) {
            vidEnv.PutBackEvent();
            _sentry.EnableClipRegion();
            propagateTo->HandleEvents();
            _sentry.DisableClipRegion();
            bPress = true;
          }
        }
        else if(vidEnv.EventMouseMove()) {
          // Propagate the mouse event
          if(propagateTo && !MouseOver(vidEnv.GetMouseX(), vidEnv.GetMouseY())) {
            vidEnv.PutBackEvent();
            _sentry.EnableClipRegion();
            propagateTo->HandleEvents();
            _sentry.DisableClipRegion();
          }
        }
        else if(vidEnv.EventButtonRelease() && !MouseOver(vidEnv.GetButtonX(), vidEnv.GetButtonY())) {
          eLEM = true;
        }
        // Close window event
        else if(vidEnv.EventCloseWindow()) {
          eLEM = true;
        }
        // Check if we should leave the editing mode
        // (as if we pressed the enter key)
        if(eLEM) {
          // Set flag
          eLEM = false;
          // Propagate the event
          if(propagateTo) {
            vidEnv.PutBackEvent();
            vidEnv.PopEvent(&eventStorage);
          }
          // Execute the confirmation function
          CharT zero = 0;
          bool  ret  = _confirmAction(ETBE_ConfirmOK, lines, curRow, curCol, NULL, zero, VKS_ENTER);
          // Exit if the confirmation is accepted
          if(ret) {
            retVal = true;
            break;
          }
          else {
            vidEnv.FreeStoredEvent(&eventStorage);
            if(bPress) vStorage.RestoreCurrentView();
            bPress        = false;
            updateDisplay = true;
          }
          // Set flags
          ctrl  = false;
          kyRep = false; kyCnt = 0;
          chAdd = 0;     eKSym = VKS_UNKNOWN;
        }
      }
      // Edit the text
      if(eKSym == VKS_LEFT) {
        if(!kyCnt) {
          curCnt = 999;
          if(curCol > 0) {
            // Move to the previous word
            if(ctrl) {
              size_t i = curCol - 1;
              if(isspace(lines[curRow][i])) 
                while(i > 0) { if(!isspace(lines[curRow][i])) break; i--; }
              while(i > 0) { if( isspace(lines[curRow][i])) break; i--; }
              curCol = i ? (i + 1) : 0;
            }
            // Move to the previous character
            else
              curCol--;
          }
          else if(curRow > 0) {
            curRow--;
            curCol = lines[curRow].length();
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_RIGHT) {
        if(!kyCnt) {
          curCnt = 999;
          if(curCol < lines[curRow].length())  {
            // Move to the next word
            if(ctrl) {
              size_t i = curCol;
              while(i < lines[curRow].length()) { if( isspace(lines[curRow][i])) break; i++; }
              while(i < lines[curRow].length()) { if(!isspace(lines[curRow][i])) break; i++; }
              curCol = i;
            }
            // Move to the next character
            else
              curCol++;
          }
          else if(curRow < (lines.size() - 1)) {
            curRow++;
            curCol = 0;
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_UP) {
        if(!kyCnt) {
          curCnt = 999;
          if(curRow > 0) {
#ifndef __USE_UTF8_DIALOG_STRING__
            if(curCol) _AsciiAdjustColPos_(curCol, vidEnv._tX11Style.ptFont, lines[curRow], lines[curRow - 1]);
#else
            if(curCol) _Utf32AdjustColPos_(curCol, vidEnv._tX11Style.ptFontSet, lines[curRow], lines[curRow - 1]);
#endif
            curRow--;
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_DOWN) {
        if(!kyCnt) {
          curCnt = 999;
          if(curRow < (lines.size() - 1)) {
#ifndef __USE_UTF8_DIALOG_STRING__
            if(curCol) _AsciiAdjustColPos_(curCol, vidEnv._tX11Style.ptFont, lines[curRow], lines[curRow + 1]);
#else
            if(curCol) _Utf32AdjustColPos_(curCol, vidEnv._tX11Style.ptFontSet, lines[curRow], lines[curRow + 1]);
#endif
            curRow++;
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_PGUP) {
        if(!kyCnt) {
          curCnt = 999;
          const size_t oldRow = curRow;
          curRow = (curRow > pgMv) ? (curRow - pgMv) : 0;
          const size_t newRow = curRow;
#ifndef __USE_UTF8_DIALOG_STRING__
            if(curCol && oldRow!=newRow) _AsciiAdjustColPos_(curCol, vidEnv._tX11Style.ptFont, lines[oldRow], lines[newRow]);
#else
            if(curCol && oldRow!=newRow) _Utf32AdjustColPos_(curCol, vidEnv._tX11Style.ptFontSet, lines[oldRow], lines[newRow]);
#endif
        }
        kyCnt++;
      }
      else if(eKSym == VKS_PGDN) {
        if(!kyCnt) {
          curCnt = 999;
          const size_t oldRow = curRow;
          curRow = (curRow < (lines.size() - pgMv)) ? (curRow + pgMv) : (lines.size() - 1);
          const size_t newRow = curRow;
#ifndef __USE_UTF8_DIALOG_STRING__
            if(curCol && oldRow!=newRow) _AsciiAdjustColPos_(curCol, vidEnv._tX11Style.ptFont, lines[oldRow], lines[newRow]);
#else
            if(curCol && oldRow!=newRow) _Utf32AdjustColPos_(curCol, vidEnv._tX11Style.ptFontSet, lines[oldRow], lines[newRow]);
#endif
        }
        kyCnt++;
      }
      else if(eKSym == VKS_ENTER) {
        if(!kyCnt) {
          CharT nl = '\n';
          // At the beginning of a line
          if(curCol == 0) {
            if(_confirmAction(ETBE_AddChar, lines, curRow, curCol, NULL, nl, VKS_ENTER)) {
              lines.insert(lines.begin() + curRow, StringT());
              curRow++;
              updateDisplay = true;
            }
          }
          // At the end of a line
          else if(curCol == lines[curRow].length()) {
            if(_confirmAction(ETBE_AddChar, lines, curRow, curCol, NULL, nl, VKS_ENTER)) {
              lines.insert(lines.begin() + curRow + 1, StringT());
              curRow++;
              updateDisplay = true;
            }
          }
          // In the middle of a line
          else if(_confirmAction(ETBE_AddChar, lines, curRow, curCol, NULL, nl, VKS_ENTER)) {
            const StringT str1 = lines[curRow].substr(0, curCol);
            const StringT str2 = lines[curRow].substr(curCol, string::npos);
            lines.erase(lines.begin() + curRow);
            lines.insert(lines.begin() + curRow    , str1);
            lines.insert(lines.begin() + curRow + 1, str2);
            curRow++;
            curCol = 0;
            updateDisplay = true;
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_BS) {
        if(!kyCnt) {
          // Not at the beginning of a line
          if(curCol > 0) {
            if(_confirmAction(ETBE_DelChar, lines, curRow, curCol - 1, NULL, lines[curRow][curCol - 1], VKS_BS)) {
              lines[curRow].erase(curCol - 1, 1);
              curCol--;
              updateDisplay = true;
            }
          }
          // At the beginning of a line
          else {
            CharT nl = '\n';
            // This is the 1st line
            if(curRow == 0) {
              // Nothing to do
            }
            // This is the last line
            else if(curRow == (lines.size() - 1)) {
              if(_confirmAction(ETBE_DelChar, lines, curRow - 1, lines[curRow - 1].length(), NULL, nl, VKS_BS)) {
                const size_t pos = lines[curRow - 1].length();
                lines[curRow - 1] += lines[curRow];
                lines.pop_back();
                curRow--;
                curCol = pos;
                updateDisplay = true;
              }
            }
            // Other lines
            else if(_confirmAction(ETBE_DelChar, lines, curRow - 1, lines[curRow - 1].length(), NULL, nl, VKS_BS)) {
              const size_t  pos = lines[curRow - 1].length();
              const StringT str = lines[curRow - 1] + lines[curRow];
              vector<StringT>::iterator it = lines.begin() + curRow;
              lines.erase(it - 1, it + 1);
              lines.insert(lines.begin() + curRow - 1, str);
              curRow--;
              curCol = pos;
              updateDisplay = true;
            }
          }
        }
        kyCnt++;
      }
      else if(eKSym == VKS_DEL) {
        if(!kyCnt) {
          // Not at the end of a line
          if(curCol < lines[curRow].length()) {
            if(_confirmAction(ETBE_DelChar, lines, curRow, curCol, NULL, lines[curRow][curCol], VKS_DEL)) {
              lines[curRow].erase(curCol, 1);
              updateDisplay = true;
            }
          }
          // At the end of a line
          else {
            CharT nl = '\n';
            // This is the last line
            if(curRow == (lines.size() - 1)) {
              // Nothing to do
            }
            // Other lines
            else if(_confirmAction(ETBE_DelChar, lines, curRow, curCol, NULL, nl, VKS_DEL)) {
              const StringT str = lines[curRow] += lines[curRow + 1];
              lines.erase(lines.begin() + curRow + 1);
              updateDisplay = true;
            }
          }
        }
        kyCnt++;
      }
      else if(chAdd) {
        if(!kyCnt) {
          // Overwrite the character in the current position
          if(overwrite) {
            if(_confirmAction(ETBE_ReplaceChar, lines, curRow, curCol, NULL, chAdd, eKSym)) {
              if(curCol < lines[curRow].length()) lines[curRow].erase(curCol, 1);
              lines[curRow].insert(curCol, 1, chAdd);
              curCol++;
              updateDisplay = true;
            }
          }
          // Append character at the current position
          else if(_confirmAction(ETBE_AddChar, lines, curRow, curCol, NULL, chAdd, eKSym)) {
            lines[curRow].insert(curCol, 1, chAdd);
            curCol++;
            updateDisplay = true;
          }
        }
        kyCnt++;
      }
      // Check the key repeat delay counter
      if(kyRep) {
        if(kyCnt >= KEYREP_DELAY) kyCnt = 0;
      } 
      else if(kyCnt >= KEYREP_START) {
        kyRep = true;
        kyCnt = 0;
      }
    }
  }
  catch(...) {
    vStorage.RestoreCurrentView();
    throw;
  }

  // Restore the view
  vStorage.RestoreCurrentView();

  // Restore the lines back to single string
  if(retVal) _MLTBComposeLines_(_csBuffer, lines);

  // Propagate the pending event (if any)
  if(eventStorage) {
    vidEnv.PushEvent(&eventStorage);
    //propagateTo->HandleEvents();
  }

  // Return the result
  return(retVal);
}

void CMLTextBox::DrawControl() const
{
  DrawControl(false);
}

void CMLTextBox::DrawControl(bool bPressed) const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Don't redraw if in editing mode
  if(_inDoEdit) return;

  // Draw face
  vidEnv.SetFillStyle(FILL_SOLID);
  if(_readOnly) vidEnv.SetForegroundColor(vidEnv._tX11Style.xFaceLocked);
  else          vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tFace.x, _tFace.y, _tFace.width, _tFace.height);

  // Draw outline
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tOutline.x, _tOutline.y, _tOutline.width, _tOutline.height);

  // Draw border
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegTL), 2);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegBR), 2);

  // Set clip region
  XRectangle rClip = _tFace;
  rClip.width  -= 2;
  rClip.height -= 2;
  XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &rClip, 1, Unsorted);

  // Prepare the GC for drawing text
  if(_tBD.bEnabled) {
    if(bPressed)
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xHotPress);
    else
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  }

  // For convenience when drawing text
  const string           &str    = _csBuffer;
  const string::size_type strLen = str.length();

  // Calculate starting height using dummy text
#ifndef __USE_UTF8_DIALOG_STRING__
  int         dr, as, ds;
  XCharStruct cs;
  vidEnv.SetFont(vidEnv._tX11Style.ptFont);
  XTextExtents(*vidEnv._tX11Style.ptFont, _DTM, strlen(_DTM), &dr, &as, &ds, &cs);
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, _DTM, strlen(_DTM), NULL, &tl);
#endif

  // Print all the text lines
  string::size_type startIndex = 0;
  string::size_type endIndex   = 0;
  ssize_t           x          = _un32Xt;
  ssize_t           y          = _un32Yt;
  while(startIndex < strLen) {
    // Get the current line text
    string subStr;
    if((endIndex = str.find('\n', startIndex)) == string::npos) {
      subStr     = str.substr(startIndex, strLen - startIndex);
      startIndex = strLen;
    }
    else {
      subStr     = str.substr(startIndex, endIndex - startIndex);
      startIndex = endIndex + 1;
    }
    // Print the text
    if(!subStr.empty()) {
#ifndef __USE_UTF8_DIALOG_STRING__
      XTextExtents(*vidEnv._tX11Style.ptFont, subStr.data(), subStr.length(), &dr, &as, &ds, &cs);
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  x, y + as, subStr.data(), subStr.length());
      y += (as + ds);
#else
      Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, subStr.data(), subStr.length(), NULL, &tl);
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      x, y - tl.y, subStr.data(), subStr.length());
      y += tl.height;
#endif
    }
    else {
#ifndef __USE_UTF8_DIALOG_STRING__
      y += (as + ds);
#else
      y += tl.height;
#endif
    }
  }

  // Restore clip region
  XRectangle clip;
  clip.x = 0; clip.y = 0; clip.width = vidEnv._tX11.un32Width; clip.height = vidEnv._tX11.un32Height;
  XSetClipRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainGC, 0, 0, &clip, 1, Unsorted);
}

void CMLTextBox::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  DrawControl(MouseOver(n32MouseX, n32MouseY) & bPressed);
}

} // namespace vge
